import threading
from typing import Any, Dict, List

from supervisely.api.api import Api
from supervisely.api.dataset_api import DatasetInfo
from supervisely.api.project_api import ProjectInfo
from supervisely.api.video.video_api import VideoInfo
from supervisely.app.widgets import (
    Button,
    Card,
    Container,
    FastTable,
    OneOf,
    RadioGroup,
    SelectDatasetTree,
    Text,
)
from supervisely.app.widgets.widget import Widget
from supervisely.project.project import ProjectType


class InputSelector:
    title = "Input data"
    description = "Select input data on which to run model for prediction"
    lock_message = None

    def __init__(self, workspace_id: int, api: Api):
        # Init Step
        self.workspace_id = workspace_id
        self.api = api
        self.display_widgets: List[Any] = []
        # -------------------------------- #

        # Init Base Widgets
        self.validator_text = None
        self.button = None
        self.container = None
        self.card = None
        # -------------------------------- #

        # Init Step Widgets
        # Images
        self.select_dataset_for_images = None
        self.select_image_container = None
        # Videos
        self.select_dataset_for_video = None
        self.select_video = None
        self.select_video_container = None
        # Selector
        self.radio = None
        self.one_of = None
        # -------------------------------- #

        # Images
        self.select_dataset_for_images = SelectDatasetTree(
            multiselect=True,
            flat=True,
            select_all_datasets=True,
            allowed_project_types=[ProjectType.IMAGES],
            always_open=False,
            compact=False,
            team_is_selectable=False,
            workspace_is_selectable=False,
            show_select_all_datasets_checkbox=True,
        )
        self.select_image_container = Container(widgets=[self.select_dataset_for_images])
        self._radio_item_images = RadioGroup.Item(
            ProjectType.IMAGES.value, "Images", content=self.select_image_container
        )
        # -------------------------------- #

        # Videos
        self.select_dataset_for_video = SelectDatasetTree(
            multiselect=True,
            flat=True,
            select_all_datasets=True,
            allowed_project_types=[ProjectType.VIDEOS],
            always_open=False,
            compact=False,
            team_is_selectable=False,
            workspace_is_selectable=False,
            show_select_all_datasets_checkbox=True,
        )
        self._video_table_columns = [
            "Video id",
            "Video name",
            "Size",
            "Duration",
            "FPS",
            "Frames count",
            "Dataset name",
            "Dataset id",
        ]
        self.select_video = FastTable(
            columns=self._video_table_columns,
            is_selectable=True,
        )
        self.select_video.hide()
        self.select_video_container = Container(
            widgets=[self.select_dataset_for_video, self.select_video]
        )
        self._radio_item_videos = RadioGroup.Item(
            ProjectType.VIDEOS.value, "Videos", content=self.select_video_container
        )
        # -------------------------------- #

        # Data type Radio Selector
        self.radio = RadioGroup(items=[self._radio_item_images, self._radio_item_videos])
        # self.radio = RadioGroup(items=[self._radio_item_images])
        # self.radio.hide()
        self.one_of = OneOf(conditional_widget=self.radio)
        # Add widgets to display ------------ #
        self.display_widgets.extend([self.radio, self.one_of])
        # ----------------------------------- #

        # Base Widgets
        self.validator_text = Text("")
        self.validator_text.hide()
        self.button = Button("Select")
        # Add widgets to display ------------ #
        self.display_widgets.extend([self.validator_text, self.button])
        # ----------------------------------- #

        # Card Layout
        self.container = Container(self.display_widgets)
        self.card = Card(
            title=self.title,
            description=self.description,
            content=self.container,
            lock_message=self.lock_message,
        )
        # ----------------------------------- #

        self._refresh_table_lock = threading.Lock()
        self._refresh_table_thread: threading.Thread = None
        self._refresh_called = False

        @self.radio.value_changed
        def input_selector_type_changed(value: str):
            self.validator_text.hide()

        @self.select_dataset_for_images.project_changed
        def _images_project_changed(project_id):
            self.validator_text.hide()

        @self.select_dataset_for_images.value_changed
        def _images_dataset_changed(dataset_ids):
            self.validator_text.hide()

        @self.select_dataset_for_video.project_changed
        def _videos_project_changed(project_id: int):
            self._refresh_video_table_called()

        @self.select_dataset_for_video.value_changed
        def _videos_dataset_changed(datasets_ids):
            self._refresh_video_table_called()

    def _refresh_video_table_called(self):
        with self._refresh_table_lock:
            self._refresh_called = True
            if self._refresh_table_thread is None or not self._refresh_table_thread.is_alive():
                self._refresh_table_thread = threading.Thread(target=self._refresh_video_table_loop)
        if self._refresh_table_thread is not None and not self._refresh_table_thread.is_alive():
            self._refresh_table_thread.start()

    def _refresh_video_table_loop(self):
        while self._refresh_called:
            with self._refresh_table_lock:
                self._refresh_called = False
            self.select_video.loading = True
            self._refresh_video_table()
            if not self._refresh_called:
                self.select_video.loading = False

    def _refresh_video_table(self):
        self.validator_text.hide()
        self.select_video.clear()
        selected_datasets = self.select_dataset_for_video.get_selected_ids()
        if not selected_datasets:
            self.select_video.hide()
        else:
            rows = []
            self.select_video.show()
            for dataset_id in selected_datasets:
                dataset_info = self.api.dataset.get_info_by_id(dataset_id)
                videos = self.api.video.get_list(dataset_id)
                for video in videos:
                    size = f"{video.frame_height}x{video.frame_width}"
                    try:
                        frame_rate = int(video.frames_count / video.duration)
                    except:
                        frame_rate = "N/A"
                    rows.append(
                        [
                            video.id,
                            video.name,
                            size,
                            video.duration,
                            frame_rate,
                            video.frames_count,
                            dataset_info.name,
                            dataset_info.id,
                        ]
                    )

            self.select_video.add_rows(rows)

    def select_project(self, project_id: int, project_info: ProjectInfo = None):
        if project_info is None:
            project_info = self.api.project.get_info_by_id(project_id)
        if project_info.type == ProjectType.IMAGES.value:
            self.select_dataset_for_images.set_project_id(project_id)
            self.select_dataset_for_images.select_all()
            self.radio.set_value(ProjectType.IMAGES.value)
        elif project_info.type == ProjectType.VIDEOS.value:
            self.select_dataset_for_video.set_project_id(project_id)
            self.select_dataset_for_video.select_all()
            self._refresh_video_table()
            self.select_video.select_rows(list(range(len(self.select_video._rows_total))))
            self.radio.set_value(ProjectType.VIDEOS.value)
        else:
            raise ValueError(f"Project of type {project_info.type} is not supported.")

    def select_datasets(self, dataset_ids: List[int], dataset_infos: List[DatasetInfo] = None):
        if dataset_infos is None:
            dataset_infos = [self.api.dataset.get_info_by_id(ds_id) for ds_id in dataset_ids]
        project_ids = set(ds.project_id for ds in dataset_infos)
        if len(project_ids) > 1:
            raise ValueError("Cannot select datasets from different projects")
        project_id = project_ids.pop()
        project_info = self.api.project.get_info_by_id(project_id)
        if project_info.type == ProjectType.IMAGES.value:
            self.select_dataset_for_images.set_project_id(project_id)
            self.select_dataset_for_images.set_dataset_ids(dataset_ids)
            self.radio.set_value(ProjectType.IMAGES.value)
        elif project_info.type == ProjectType.VIDEOS.value:
            self.select_dataset_for_video.set_project_id(project_id)
            self.select_dataset_for_video.set_dataset_ids(dataset_ids)
            self._refresh_video_table()
            self.select_video.select_rows(list(range(self.select_video._rows_total)))
            self.radio.set_value(ProjectType.VIDEOS.value)
        else:
            raise ValueError(f"Project of type {project_info.type} is not supported.")

    def select_videos(self, video_ids: List[int], video_infos: List[VideoInfo] = None):
        if video_infos is None:
            video_infos = self.api.video.get_info_by_id_batch(video_ids)
        project_id = video_infos[0].project_id
        self.select_dataset_for_video.set_project_id(project_id)
        self.select_dataset_for_video.select_all()
        self._refresh_video_table()
        self.select_video.select_row_by_value("id", video_ids)
        self.radio.set_value(ProjectType.VIDEOS.value)

    def disable(self):
        for widget in self.widgets_to_disable:
            widget.disable()

    def enable(self):
        for widget in self.widgets_to_disable:
            widget.enable()

    @property
    def widgets_to_disable(self) -> List[Widget]:
        return [
            # Images Selector
            self.select_dataset_for_images,
            self.select_dataset_for_images._select_project,
            self.select_dataset_for_images._select_dataset,
            # Videos Selector
            self.select_dataset_for_video,
            self.select_dataset_for_video._select_project,
            self.select_dataset_for_video._select_dataset,
            self.select_video,
            # Controls
            self.radio,
            self.one_of,
        ]

    def get_settings(self) -> Dict[str, Any]:
        if self.radio.get_value() == ProjectType.IMAGES.value:
            return {
                "project_id": self.select_dataset_for_images.get_selected_project_id(),
                "dataset_ids": self.select_dataset_for_images.get_selected_ids(),
            }
        if self.radio.get_value() == ProjectType.VIDEOS.value:
            rows = self.select_video.get_selected_rows()
            if rows:
                video_ids = [row.row[0] for row in rows]
            else:
                video_ids = None
            return {"video_ids": video_ids}

    def load_from_json(self, data):
        if "video_ids" in data:
            video_ids = data["video_ids"]
            if not video_ids:
                raise ValueError("Video ids cannot be empty")
            video_infos = self.api.video.get_info_by_id_batch(video_ids)
            if not video_infos:
                raise ValueError(f"Videos with video ids {video_ids} are not found")
            self.select_videos(video_ids, video_infos)
        elif "dataset_ids" in data:
            dataset_ids = data["dataset_ids"]
            self.select_datasets(dataset_ids)
        elif "project_id" in data:
            project_id = data["project_id"]
            self.select_project(project_id)

    def get_project_id(self) -> int:
        if self.radio.get_value() == ProjectType.IMAGES.value:
            return self.select_dataset_for_images.project_id
        if self.radio.get_value() == ProjectType.VIDEOS.value:
            return self.select_dataset_for_video.project_id
        return None

    def validate_step(self) -> bool:
        self.validator_text.hide()
        if self.radio.get_value() == ProjectType.IMAGES.value:
            selected_ids = self.select_dataset_for_images.get_selected_ids()
            if selected_ids is None:
                self.validator_text.set(text="Select a project", status="error")
                self.validator_text.show()
                return False
            if len(selected_ids) == 0:
                self.validator_text.set(text="Select at least one dataset", status="error")
                self.validator_text.show()
                return False
        if self.radio.get_value() == ProjectType.VIDEOS.value:
            if not self.select_dataset_for_video.get_selected_ids():
                self.validator_text.set(text="Select a dataset", status="error")
                self.validator_text.show()
                return False
            if self.select_video._rows_total == 0:
                self.validator_text.set(
                    text="No videos found in the selected dataset", status="error"
                )
                self.validator_text.show()
                return False
            if self.select_video.get_selected_rows() == []:
                self.validator_text.set(text="Select a video", status="error")
                self.validator_text.show()
                return False
        return True



import copy
import traceback
from typing import Any, Callable, Dict, List, Literal, Optional, Union

import numpy as np
import pandas as pd

from supervisely.app import DataJson
from supervisely.app.content import StateJson
from supervisely.app.widgets import Widget
from supervisely.app.widgets_context import JinjaWidgets
from supervisely.project.project_meta import ProjectMeta
from supervisely.sly_logger import logger


class EventLinstenerError(Exception):
    def __init__(self, message="An exception occurred due to conflicting event listeners."):
        self.message = message
        super().__init__(self.message)


class FastTable(Widget):
    """FastTable widget in Supervisely allows for displaying and manipulating data of various
    dataset statistics and processing it on the server side.

    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/tables/fasttable>`_
        (including screenshots and examples).

    :param data: dataset or project data in different formats:
    :type data: Union[pd.DataFrame, List]
    :param columns: List of column names
    :type columns: List, optional
    :param columns_options: List of dicts with options for each column
    :type columns_options: List[dict], optional
    :param project_meta: Project meta information
    :type project_meta: Union[ProjectMeta, dict], optional
    :param fixed_columns: Number of fixed columns
    :type fixed_columns: Literal[1], optional
    :param page_size: Number of rows per page
    :type page_size: int, optional
    :param sort_column_idx: Index of the column to sort by
    :type sort_column_idx: int, optional
    :param sort_order: Sorting order
    :type sort_order: Literal["asc", "desc"], optional
    :param width: Width of the widget, e.g. "200px", or "100%"
    :type width: str, optional
    :param widget_id: Unique widget identifier.
    :type widget_id: str, optional
    :param show_header: Whether to show table header
    :type show_header: bool, optional
    :param is_radio: Enable radio button selection mode (single row selection)
    :type is_radio: bool, optional
    :param is_selectable: Enable multiple row selection
    :type is_selectable: bool, optional
    :param header_left_content: Widget to display in the left side of the header
    :type header_left_content: Widget, optional
    :param header_right_content: Widget to display in the right side of the header
    :type header_right_content: Widget, optional
    :param max_selected_rows: Maximum number of rows that can be selected
    :type max_selected_rows: int, optional
    :param search_position: Position of the search input ("left" or "right")
    :type search_position: Literal["left", "right"], optional


    :Usage example:
    .. code-block:: python

        from supervisely.app.widgets import FastTable

        data = [["apple", "21"], ["banana", "15"]]
        columns = ["Class", "Items"]
        dataframe = pd.DataFrame(data=data, columns=columns)
        columns_options = [
            { "type": "class"},
            { "maxValue": 21, "postfix": "pcs", "tooltip": "description text", "subtitle": "boxes" }
        ]

        meta_path = "meta.json"  # define file path
        with open(meta_path, "r") as json_file:
            meta = json.load(json_file)

        fast_table = FastTable(
            data=dataframe,
            project_meta=meta,
            columns_options=columns_options,
        )
    """

    class Routes:
        SELECTION_CHANGED = "selection_changed_cb"
        ROW_CLICKED = "row_clicked_cb"
        CELL_CLICKED = "cell_clicked_cb"
        UPDATE_DATA = "update_data_cb"

    class ClickedRow:
        def __init__(
            self,
            row: List,
            row_index: int = None,
        ):
            self.row = row
            self.row_index = row_index

    class ClickedCell:
        def __init__(
            self,
            row: List,
            column_index: int,
            row_index: int = None,
            column_name: int = None,
            column_value: int = None,
        ):
            self.row = row
            self.column_index = column_index
            self.row_index = row_index
            self.column_name = column_name
            self.column_value = column_value

    class ColumnData:
        def __init__(self, name, is_widget=False, widget: Widget = None):
            self.name = name
            self.is_widget = is_widget
            self.widget = widget

        @property
        def widget_html(self):
            html = self.widget.to_html()
            html = html.replace(f".{self.widget.widget_id}", "[JSON.parse(cellValue).widget_id]")
            html = html.replace(
                f"/{self.widget.widget_id}", "/' + JSON.parse(cellValue).widget_id + '"
            )
            if hasattr(self.widget, "_widgets"):
                for i, widget in enumerate(self.widget._widgets):
                    html = html.replace(
                        f".{widget.widget_id}", f"[JSON.parse(cellValue).widgets[{i}]]"
                    )
                    html = html.replace(
                        f"/{widget.widget_id}", f"/' + JSON.parse(cellValue).widgets[{i}] + '"
                    )
            return html

    def __init__(
        self,
        data: Optional[Union[pd.DataFrame, List]] = None,
        columns: Optional[List] = None,
        columns_options: Optional[List[dict]] = None,
        project_meta: Optional[Union[ProjectMeta, dict]] = None,
        fixed_columns: Optional[Literal[1]] = None,
        page_size: Optional[int] = 10,
        sort_column_idx: int = None,
        sort_order: Optional[Literal["asc", "desc"]] = None,
        width: Optional[str] = "auto",
        widget_id: Optional[str] = None,
        show_header: bool = True,
        is_radio: bool = False,
        is_selectable: bool = False,
        header_left_content: Optional[Widget] = None,
        header_right_content: Optional[Widget] = None,
        max_selected_rows: Optional[int] = None,
        search_position: Optional[Literal["left", "right"]] = None,
    ):
        self._supported_types = tuple([pd.DataFrame, list, type(None)])
        self._row_click_handled = False
        self._cell_click_handled = False
        self._selection_changed_handled = False
        self._columns = columns
        self._columns_data = []
        if columns is None:
            self._columns_first_idx = None
        else:
            self._columns_first_idx = []
            for col in columns:
                if isinstance(col, str):
                    self._columns_first_idx.append(col)
                    self._columns_data.append(self.ColumnData(name=col))
                elif isinstance(col, tuple):
                    self._columns_first_idx.append(col[0])
                    self._columns_data.append(
                        self.ColumnData(name=col[0], is_widget=True, widget=col[1])
                    )
                else:
                    raise TypeError(f"Column name must be a string or a tuple, got {type(col)}")

        self._columns_options = columns_options
        self._sorted_data = None
        self._filtered_data = None
        self._searched_data = None
        self._active_page = 1
        self._width = width
        self._selected_rows = []
        self._selected_cell = None
        self._clicked_row = None
        self._is_row_clickable = False
        self._is_cell_clickable = False
        self._search_str = ""
        self._show_header = show_header
        self._project_meta = self._unpack_project_meta(project_meta)
        self._header_left_content = header_left_content
        self._header_right_content = header_right_content
        self._max_selected_rows = max_selected_rows
        acceptable_search_positions = ["left", "right"]
        self._search_position = search_position if search_position in acceptable_search_positions else "left"

        # table_options
        self._page_size = page_size
        self._fix_columns = self._validate_fix_columns_value(fixed_columns)
        self._sort_column_idx = sort_column_idx
        self._sort_order = sort_order
        self._validate_sort_attrs()
        self._is_radio = is_radio
        self._is_selectable = is_selectable
        self._search_function = self._default_search_function
        self._sort_function = self._default_sort_function
        self._filter_function = self._default_filter_function
        self._filter_value = None

        # to avoid errors with the duplicated names in columns
        self._multi_idx_columns = None

        # prepare source_data
        self._validate_input_data(data)
        self._source_data = self._prepare_input_data(data)

        # prepare parsed_source_data, sliced_data, parsed_active_data
        (
            self._parsed_source_data,
            self._sliced_data,
            self._parsed_active_data,
        ) = self._prepare_working_data()

        self._rows_total = len(self._parsed_source_data["data"])

        if self._is_radio and self._rows_total > 0:
            self._selected_rows = [self._parsed_source_data["data"][0]]

        super().__init__(widget_id=widget_id, file_path=__file__)

        script_path = "./sly/css/app/widgets/fast_table/script.js"
        JinjaWidgets().context["__widget_scripts__"][self.__class__.__name__] = script_path

        filter_changed_route_path = self.get_route_path(FastTable.Routes.UPDATE_DATA)
        server = self._sly_app.get_server()

        @server.post(filter_changed_route_path)
        def _filter_changed_handler():
            self._refresh()

    def _refresh(self):
        # TODO sort widgets
        self._active_page = StateJson()[self.widget_id]["page"]
        self._sort_order = StateJson()[self.widget_id]["sort"]["order"]
        self._sort_column_idx = StateJson()[self.widget_id]["sort"]["column"]
        search_value = StateJson()[self.widget_id]["search"]
        self._filtered_data = self._filter(self._filter_value)
        self._searched_data = self._search(search_value)
        self._rows_total = len(self._searched_data)

        # if active page is greater than the number of pages (e.g. after filtering)
        max_page = (self._rows_total - 1) // self._page_size + 1
        if (self._rows_total > 0 and self._active_page == 0) or self._active_page > max_page:
            self._active_page = 1
            StateJson()[self.widget_id]["page"] = self._active_page

        self._sorted_data = self._sort_table_data(self._searched_data)
        self._sliced_data = self._slice_table_data(self._sorted_data, actual_page=self._active_page)
        self._parsed_active_data = self._unpack_pandas_table_data(self._sliced_data)
        StateJson().send_changes()
        DataJson()[self.widget_id]["data"] = {
            i: row for i, row in enumerate(self._parsed_active_data["data"])
        }
        DataJson()[self.widget_id]["total"] = self._rows_total
        DataJson().send_changes()
        StateJson()["reactToChanges"] = True
        StateJson().send_changes()

    def get_json_data(self) -> Dict[str, Any]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.
        Dictionary contains the following fields:
            - data: table data
            - columns: list of column names
            - projectMeta: project meta information
            - columnsOptions: list of dicts with options for each column
            - total: total number of rows
            - options: table options with the following fields:
                - isRowClickable: whether rows are clickable
                - isCellClickable: whether cells are clickable
                - fixColumns: number of fixed columns
                - isRadio: whether radio button selection mode is enabled
                - isRowSelectable: whether multiple row selection is enabled
                - maxSelectedRows: maximum number of rows that can be selected
                - searchPosition: position of the search input ("left" or "right")
            - pageSize: number of rows per page
            - showHeader: whether to show table header
            - selectionChangedHandled: whether selection changed event listener is set

        :return: Dictionary with widget data
        :rtype: Dict[str, Any]
        """
        return {
            "data": {i: row for i, row in enumerate(self._parsed_active_data["data"])},
            "columns": self._parsed_source_data["columns"],
            "projectMeta": self._project_meta,
            "columnsOptions": self._columns_options,
            "total": self._rows_total,
            "options": {
                "isRowClickable": self._is_row_clickable,
                "isCellClickable": self._is_cell_clickable,
                "fixColumns": self._fix_columns,
                "isRadio": self._is_radio,
                "isRowSelectable": self._is_selectable,
                "maxSelectedRows": self._max_selected_rows,
                "searchPosition": self._search_position,
            },
            "pageSize": self._page_size,
            "showHeader": self._show_header,
            "selectionChangedHandled": self._selection_changed_handled,
        }

    def get_json_state(self) -> Dict[str, Any]:
        """Returns dictionary with widget state.
        Dictionary contains the following fields:
            - search: search string
            - selectedRows: selected rows
            - selectedCell: selected cell
            - clickedRow: clicked row
            - page: active page
            - sort: sorting options with the following fields:
                - column: index of the column to sort by
                - order: sorting order

        :return: Dictionary with widget state
        :rtype: Dict[str, Any]
        """
        return {
            "search": self._search_str,
            "selectedRows": self._selected_rows,
            "selectedCell": self._selected_cell,
            "clickedRow": self._clicked_row,
            "page": self._active_page,
            "sort": {
                "column": self._sort_column_idx,
                "order": self._sort_order,
            },
        }

    @property
    def fixed_columns_num(self) -> int:
        """Returns number of fixed columns.

        :return: Number of fixed columns
        :rtype: int
        """
        return self._fix_columns

    @fixed_columns_num.setter
    def fixed_columns_num(self, value: int) -> None:
        """Sets number of fixed columns.

        :param value: Number of fixed columns
        :type value: int
        """
        self._fix_columns = self._validate_fix_columns_value(value)
        DataJson()[self.widget_id]["options"]["fixColumns"] = self._fix_columns

    @property
    def project_meta(self) -> Dict[str, Any]:
        """Returns project meta information.

        :return: Project meta information
        :rtype: Dict[str, Any]
        """
        return self._project_meta

    @project_meta.setter
    def project_meta(self, meta: Union[ProjectMeta, Dict]) -> None:
        """Sets project meta information.

        :param meta: Project meta information
        :type meta: Union[ProjectMeta, Dict]
        """
        self._project_meta = self._unpack_project_meta(meta)
        DataJson()[self.widget_id]["projectMeta"] = self._project_meta

    @property
    def page_size(self) -> int:
        return self._page_size

    @page_size.setter
    def page_size(self, size: int):
        self._page_size = size
        DataJson()[self.widget_id]["pageSize"] = self._page_size

    def set_sort(
        self, func: Callable[[pd.DataFrame, int, Optional[Literal["asc", "desc"]]], pd.DataFrame]
    ) -> None:
        """Sets custom sort function for the table.

        :param func: Custom sort function
        :type func: Callable[[pd.DataFrame, int, Optional[Literal["asc", "desc"]]], pd.DataFrame]
        """
        self._sort_function = func

    def set_search(self, func: Callable[[pd.DataFrame, str], pd.DataFrame]) -> None:
        """Sets custom search function for the table.

        :param func: Custom search function
        :type func: Callable[[pd.DataFrame, str], pd.DataFrame]
        """
        self._search_function = func

    def set_filter(self, filter_function: Callable[[pd.DataFrame, Any], pd.DataFrame]) -> None:
        """Sets a custom filter function for the table.
        first argument is a DataFrame, second argument is a filter value.

        :param filter_function: Custom filter function
        :type filter_function: Callable[[pd.DataFrame, Any], pd.DataFrame]
        """
        if filter_function is None:
            filter_function = self._default_filter_function
        self._filter_function = filter_function

    def read_json(self, data: Dict, meta: Dict = None, custom_columns: Optional[List[Union[str, tuple]]] = None) -> None:
        """Replace table data with options and project meta in the widget

        :param data: Table data with options:
            - data: table data
            - columns: list of column names
            - projectMeta: project meta information - if provided
            - columnsOptions: list of dicts with options for each column
            - total: total number of rows
            - options: table options with the following fields:
                - isRowClickable: whether rows are clickable
                - isCellClickable: whether cells are clickable
                - fixColumns: number of fixed columns
                - isRadio: whether radio button selection mode is enabled
                - isRowSelectable: whether multiple row selection is enabled
                - maxSelectedRows: maximum number of rows that can be selected
                - searchPosition: position of the search input ("left" or "right")
            - pageSize: number of rows per page
            - showHeader: whether to show table header
            - selectionChangedHandled: whether selection changed event listener is set

        :type data: dict
        :param meta: Project meta information
        :type meta: dict
        :param custom_columns: List of column names. Can include widgets as tuples (column_name, widget)
        :type custom_columns: List[Union[str, tuple]], optional

        Example of data dict:
        .. code-block:: python

            data = {
                "data": [["apple", "21"], ["banana", "15"]],
                "columns": ["Class", "Items"],
                "columnsOptions": [
                    { "type": "class"},
                    { "maxValue": 21, "postfix": "pcs", "tooltip": "description text", "subtitle": "boxes" }
                ],
                "options": {
                    "isRowClickable": True,
                    "isCellClickable": True,
                    "fixColumns": 1,
                    "isRadio": False,
                    "isRowSelectable": True,
                    "maxSelectedRows": 5,
                    "searchPosition": "right",
                    "sort": {"column": 0, "order": "asc"},
                },
            }
        """
        self._columns_options = self._prepare_json_data(data, "columnsOptions")
        self._read_custom_columns(custom_columns)
        if not self._columns_first_idx:
            self._columns_first_idx = self._prepare_json_data(data, "columns")
        self._table_options = self._prepare_json_data(data, "options")
        self._project_meta = self._unpack_project_meta(meta)
        table_data = data.get("data", None)
        self._validate_input_data(table_data)
        self._source_data = self._prepare_input_data(table_data)
        (
            self._parsed_source_data,
            self._sliced_data,
            self._parsed_active_data,
        ) = self._prepare_working_data()
        self._rows_total = len(self._parsed_source_data["data"])
        init_options = DataJson()[self.widget_id]["options"]
        init_options.update(self._table_options)
        sort = init_options.pop("sort", {"column": None, "order": None})
        self._active_page = 1
        self._sort_column_idx = sort.get("column", None)
        if self._sort_column_idx is not None and self._sort_column_idx > len(self._columns_first_idx) - 1:
            self._sort_column_idx = None
        self._sort_order = sort.get("order", None)
        self._page_size = init_options.pop("pageSize", 10) 
        DataJson()[self.widget_id]["data"] = {
            i: row for i, row in enumerate(self._parsed_active_data["data"])
        }
        DataJson()[self.widget_id]["columns"] = self._parsed_active_data["columns"]
        DataJson()[self.widget_id]["columnsOptions"] = self._columns_options
        DataJson()[self.widget_id]["options"] = init_options
        DataJson()[self.widget_id]["total"] = len(self._source_data)
        DataJson()[self.widget_id]["pageSize"] = self._page_size
        DataJson()[self.widget_id]["projectMeta"] = self._project_meta
        StateJson()[self.widget_id]["sort"]["column"] = self._sort_column_idx
        StateJson()[self.widget_id]["sort"]["order"] = self._sort_order
        StateJson()[self.widget_id]["page"] = self._active_page
        StateJson()[self.widget_id]["selectedRows"] = []
        StateJson()[self.widget_id]["selectedCell"] = None
        self._maybe_update_selected_row()
        self._validate_sort_attrs()
        DataJson().send_changes()
        StateJson().send_changes()

    def read_pandas(self, data: pd.DataFrame) -> None:
        """Replace table data (rows and columns) in the widget.

        :param data: Table data
        :type data: pd.DataFrame
        """
        self._source_data = self._prepare_input_data(data)
        self._sorted_data = self._sort_table_data(self._source_data)
        self._sliced_data = self._slice_table_data(self._sorted_data)
        self._parsed_active_data = self._unpack_pandas_table_data(self._sliced_data)
        self._parsed_source_data = self._unpack_pandas_table_data(self._source_data)
        self._rows_total = len(self._parsed_source_data["data"])
        DataJson()[self.widget_id]["data"] = {
            i: row for i, row in enumerate(self._parsed_active_data["data"])
        }
        DataJson()[self.widget_id]["columns"] = self._parsed_active_data["columns"]
        DataJson()[self.widget_id]["total"] = len(self._source_data)
        DataJson().send_changes()
        self._active_page = 1
        StateJson()[self.widget_id]["page"] = self._active_page
        StateJson().send_changes()
        self.clear_selection()

    def to_json(self, active_page: Optional[bool] = False) -> Dict[str, Any]:
        """Export table data with current options as dict.

        Dictionary contains the following fields:
            - data: table data
            - columns: list of column names
            - projectMeta: project meta information - if provided
            - columnsOptions: list of dicts with options for each column
            - total: total number of rows
            - options: table options with the following fields:
                - isRowClickable: whether rows are clickable
                - isCellClickable: whether cells are clickable
                - fixColumns: number of fixed columns
                - isRadio: whether radio button selection mode is enabled
                - isRowSelectable: whether multiple row selection is enabled
                - maxSelectedRows: maximum number of rows that can be selected
                - searchPosition: position of the search input ("left" or "right")
            - pageSize: number of rows per page
            - showHeader: whether to show table header
            - selectionChangedHandled: whether selection changed event listener is set

        :param active_page: Specifies the size of the data to be exported. If True - returns only the active page of the table
        :type active_page: Optional[bool]
        :return: Table data with current options
        :rtype: dict
        """
        if active_page is True:
            temp_parsed_data = [d["items"] for d in self._parsed_active_data["data"]]
        else:
            temp_parsed_data = [d["items"] for d in self._parsed_source_data["data"]]
        widget_data = {}
        widget_data["data"] = temp_parsed_data
        widget_data["columns"] = DataJson()[self.widget_id]["columns"]
        widget_data["options"] = copy.deepcopy(DataJson()[self.widget_id]["options"])
        widget_data["columnsOptions"] = DataJson()[self.widget_id]["columnsOptions"]
        sort = copy.deepcopy(StateJson()[self.widget_id]["sort"])
        sort["columnIndex"] = sort.get("column", None)
        sort.pop("column", None)
        widget_data["options"]["sort"] = sort
        return widget_data

    def to_pandas(self, active_page=False) -> pd.DataFrame:
        """Export only table data (rows and columns) as Pandas Dataframe.

        :param active_page: Specifies the size of the data to be exported. If True - returns only the active page of the table
        :type active_page: bool
        :return: Table data
        :rtype: pd.DataFrame
        """
        if active_page is True:
            temp_parsed_data = [d["items"] for d in self._parsed_active_data["data"]]
        else:
            temp_parsed_data = [d["items"] for d in self._parsed_source_data["data"]]
        packed_data = pd.DataFrame(data=temp_parsed_data, columns=self._columns_first_idx)
        return packed_data

    def clear_selection(self) -> None:
        """Clears the selection of the table."""
        StateJson()[self.widget_id]["selectedRows"] = []
        StateJson()[self.widget_id]["selectedCell"] = None
        StateJson().send_changes()
        self._maybe_update_selected_row()

    def get_selected_row(self) -> ClickedRow:
        """Returns the selected row.

        :return: Selected row
        :rtype: ClickedRow
        """
        if self._is_radio or self._is_selectable:
            selected_rows = StateJson()[self.widget_id]["selectedRows"]
            if selected_rows is None:
                return None
            if len(selected_rows) == 0:
                return None
            if len(selected_rows) > 1:
                raise ValueError(
                    "Multiple rows selected. Use get_selected_rows() method to get all selected rows."
                )
            row = selected_rows[0]
            row_index = row["idx"]
            row = row.get("row", row.get("items", None))
            if row_index is None or row is None:
                return None
            return self.ClickedRow(row, row_index)
        return self.get_clicked_row()

    def get_selected_rows(self) -> List[ClickedRow]:
        if self._is_radio or self._is_selectable:
            selected_rows = StateJson()[self.widget_id]["selectedRows"]
            rows = []
            for row in selected_rows:
                row_index = row["idx"]
                row_data = row.get("row", row.get("items", None))
                if row_index is None or row_data is None:
                    continue
                rows.append(self.ClickedRow(row_data, row_index))
            return rows
        return [self.get_clicked_row()]

    def get_clicked_row(self) -> ClickedRow:
        clicked_row = StateJson()[self.widget_id]["clickedRow"]
        if clicked_row is None:
            return None
        row_index = clicked_row["idx"]
        row = clicked_row["row"]
        if row_index is None or row is None:
            return None
        return self.ClickedRow(row, row_index)

    def get_clicked_cell(self) -> ClickedCell:
        """Returns the selected cell.

        :return: Selected cell
        :rtype: ClickedCell
        """
        cell_data = StateJson()[self.widget_id]["clickedCell"]
        if cell_data is None:
            return None
        row_index = cell_data["idx"]
        row = cell_data["row"]
        column_index = cell_data["column"]
        column_name = self._columns_first_idx[column_index]
        column_value = row[column_index]
        if column_index is None or row is None:
            return None
        return self.ClickedCell(row, column_index, row_index, column_name, column_value)

    def get_selected_cell(self) -> ClickedCell:
        """Alias for get_clicked_cell method.
        Will be removed in future versions.
        """
        return self.get_clicked_cell()

    def _maybe_update_selected_row(self) -> None:
        if self._is_radio:
            if self._rows_total != 0:
                self.select_row(0)
            else:
                self._selected_rows = []
                StateJson()[self.widget_id]["selectedRows"] = self._selected_rows
                StateJson().send_changes()
            return
        if not self._selected_rows:
            return
        if self._rows_total == 0:
            self._selected_rows = []
            StateJson()[self.widget_id]["selectedRows"] = self._selected_rows
            StateJson().send_changes()
            return
        if self._is_selectable:
            updated_selected_rows = []
            for row in self._parsed_source_data["data"]:
                items = row.get("items", row.get("row", None))
                if items is not None:
                    for selected_row in self._selected_rows:
                        if selected_row.get("row", selected_row.get("items", None)) == items:
                            updated_selected_rows.append(row)
            self._selected_rows = updated_selected_rows
            StateJson()[self.widget_id]["selectedRows"] = self._selected_rows
            StateJson().send_changes()

    def insert_row(self, row: List, index: Optional[int] = -1) -> None:
        """Inserts a row into the table to the specified position.

        :param row: Row to insert
        :type row: List
        :param index: Index of the row to insert
        :type index: Optional[int]
        """
        self._validate_table_sizes(row)
        self._validate_row_values_types(row)
        index = len(self._source_data) if index > len(self._source_data) or index < 0 else index

        self._source_data = pd.concat(
            [
                self._source_data.iloc[:index],
                pd.DataFrame([row], columns=self._source_data.columns),
                self._source_data.iloc[index:],
            ]
        ).reset_index(drop=True)
        (
            self._parsed_source_data,
            self._sliced_data,
            self._parsed_active_data,
        ) = self._prepare_working_data()
        self._rows_total = len(self._parsed_source_data["data"])
        DataJson()[self.widget_id]["data"] = {
            i: row for i, row in enumerate(self._parsed_active_data["data"])
        }
        DataJson()[self.widget_id]["total"] = self._rows_total
        DataJson().send_changes()
        self._maybe_update_selected_row()

    def add_rows(self, rows: List):
        for row in rows:
            self._validate_table_sizes(row)
            self._validate_row_values_types(row)
        self._source_data = pd.concat(
            [self._source_data, pd.DataFrame(rows, columns=self._source_data.columns)]
        ).reset_index(drop=True)
        (
            self._parsed_source_data,
            self._sliced_data,
            self._parsed_active_data,
        ) = self._prepare_working_data()
        self._rows_total = len(self._parsed_source_data["data"])
        DataJson()[self.widget_id]["data"] = {
            i: row for i, row in enumerate(self._parsed_active_data["data"])
        }
        DataJson()[self.widget_id]["total"] = self._rows_total
        DataJson().send_changes()
        self._maybe_update_selected_row()

    def pop_row(self, index: Optional[int] = -1) -> List:
        """Removes a row from the table at the specified position and returns it.

        :param index: Index of the row to remove
        :type index: Optional[int]
        :return: Removed row
        :rtype: List
        """
        index = (
            len(self._parsed_source_data["data"]) - 1
            if index > len(self._parsed_source_data["data"]) or index < 0
            else index
        )

        if len(self._parsed_source_data["data"]) != 0:
            popped_row = self._source_data.loc[index].values
            self._source_data = self._source_data.drop(index)
            (
                self._parsed_source_data,
                self._sliced_data,
                self._parsed_active_data,
            ) = self._prepare_working_data()
            self._rows_total = len(self._parsed_source_data["data"])
            DataJson()[self.widget_id]["data"] = {
                i: row for i, row in enumerate(self._parsed_active_data["data"])
            }
            DataJson()[self.widget_id]["total"] = self._rows_total
            self._maybe_update_selected_row()
            return popped_row

    def clear(self) -> None:
        """Clears the table data."""
        self._source_data = pd.DataFrame(columns=self._columns_first_idx)
        self._parsed_source_data = {"data": [], "columns": []}
        self._sliced_data = pd.DataFrame(columns=self._columns_first_idx)
        self._parsed_active_data = {"data": [], "columns": []}
        self._rows_total = 0
        DataJson()[self.widget_id]["data"] = {}
        DataJson()[self.widget_id]["total"] = 0
        DataJson().send_changes()
        self._maybe_update_selected_row()

    def row_click(self, func: Callable[[ClickedRow], Any]) -> Callable[[], None]:
        """Decorator for function that handles row click event.

        :param func: Function that handles row click event
        :type func: Callable[[ClickedRow], Any]
        :return: Decorated function
        :rtype: Callable[[], None]
        """
        row_clicked_route_path = self.get_route_path(FastTable.Routes.ROW_CLICKED)
        server = self._sly_app.get_server()

        self._row_click_handled = True
        self._is_row_clickable = True
        DataJson()[self.widget_id]["options"]["isRowClickable"] = self._is_row_clickable
        DataJson().send_changes()

        if self._cell_click_handled is True:
            message = "An exception occurred due to conflicting event listeners: 'row_click' and 'cell_click'. To avoid errors, specify only one event listener. The 'cell_click' listener includes all information about the row as well."
            raise EventLinstenerError(message)

        @server.post(row_clicked_route_path)
        def _click():
            try:
                clicked_row = self.get_selected_row()
                if clicked_row is None:
                    return
                func(clicked_row)
            except Exception as e:
                logger.error(traceback.format_exc(), exc_info=True, extra={"exc_str": str(e)})
                raise e

        return _click

    def cell_click(self, func: Callable[[ClickedCell], Any]) -> Callable[[], None]:
        """Decorator for function that handles cell click event.

        :param func: Function that handles cell click event
        :type func: Callable[[ClickedCell], Any]
        :return: Decorated function
        :rtype: Callable[[], None]
        """
        cell_clicked_route_path = self.get_route_path(FastTable.Routes.CELL_CLICKED)
        server = self._sly_app.get_server()

        self._cell_click_handled = True
        self._is_cell_clickable = True
        DataJson()[self.widget_id]["options"]["isCellClickable"] = self._is_cell_clickable
        DataJson().send_changes()

        if self._row_click_handled is True:
            message = "An exception occurred due to conflicting event listeners: 'row_click' and 'cell_click'. To avoid errors, specify only one event listener. The 'cell_click' listener includes all information about the row as well."
            raise EventLinstenerError(message)

        @server.post(cell_clicked_route_path)
        def _click():
            try:
                clicked_cell = self.get_selected_cell()
                if clicked_cell is None:
                    return
                func(clicked_cell)
            except Exception as e:
                logger.error(traceback.format_exc(), exc_info=True, extra={"exc_str": str(e)})
                raise e

        return _click

    def _default_filter_function(self, data: pd.DataFrame, filter_value: Any) -> pd.DataFrame:
        return data

    def _filter_table_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """Filter source data using a self._filter_function as filter function.
        To apply a custom filter function, use the set_filter method.

        :return: Filtered data
        :rtype: pd.DataFrame
        """
        filtered_data = self._filter_function(data, self._filter_value)
        return filtered_data

    def _filter(self, filter_value: Any) -> pd.DataFrame:
        filtered_data = self._source_data.copy()
        if filter_value is None:
            return filtered_data
        if self._filter_value != filter_value:
            self._active_page = 1
            StateJson()[self.widget_id]["page"] = self._active_page
            StateJson().send_changes()
        self._filter_value = filter_value
        filtered_data = self._filter_table_data(filtered_data)
        return filtered_data

    def filter(self, filter_value) -> None:
        self._filter_value = filter_value
        self._refresh()

    def _default_search_function(self, data: pd.DataFrame, search_value: str) -> pd.DataFrame:
        data = data[data.applymap(lambda x: search_value in str(x)).any(axis=1)]
        return data

    def _search(self, search_value: str) -> pd.DataFrame:
        """Search source data for search value.

        :param search_value: Search value
        :type search_value: str
        :return: Filtered data
        :rtype: pd.DataFrame
        """
        filtered_data = self._filtered_data.copy()
        if search_value == "":
            return filtered_data
        if self._search_str != search_value:
            self._active_page = 1
            StateJson()[self.widget_id]["page"] = self._active_page
            StateJson().send_changes()
        filtered_data = self._search_function(filtered_data, search_value)
        self._search_str = search_value
        return filtered_data

    def search(self, search_value: str) -> None:
        StateJson()[self.widget_id]["search"] = search_value
        StateJson().send_changes()
        self._refresh()

    def _default_sort_function(
        self,
        data: pd.DataFrame,
        column_idx: Optional[int],
        order: Optional[Literal["asc", "desc"]],
    ) -> pd.DataFrame:
        if order == "asc":
            ascending = True
        else:
            ascending = False
        try:
            data = data.sort_values(by=data.columns[column_idx], ascending=ascending)
        except IndexError as e:
            e.args = (
                f"Sorting by column idx = {column_idx} is not possible, your table has only {len(data.columns)} columns with idx from 0 to {len(data.columns) - 1}",
            )
            raise e
        return data

    def sort(
        self, column_idx: Optional[int] = None, order: Optional[Literal["asc", "desc"]] = None
    ) -> None:
        """Sorts table data by column index and order.

        :param column_idx: Index of the column to sort by
        :type column_idx: Optional[int]
        :param order: Sorting order
        :type order: Optional[Literal["asc", "desc"]]
        """
        self._sort_column_idx = column_idx
        self._sort_order = order
        self._validate_sort_attrs()
        if self._sort_column_idx is not None:
            StateJson()[self.widget_id]["sort"]["column"] = self._sort_column_idx
        if self._sort_order is not None:
            StateJson()[self.widget_id]["sort"]["order"] = self._sort_order
        self._filtered_data = self._filter(self._filter_value)
        self._searched_data = self._search(self._search_str)
        self._rows_total = len(self._searched_data)
        self._sorted_data = self._sort_table_data(self._searched_data)
        self._sliced_data = self._slice_table_data(self._sorted_data, actual_page=self._active_page)
        self._parsed_active_data = self._unpack_pandas_table_data(self._sliced_data)
        DataJson()[self.widget_id]["data"] = {
            i: row for i, row in enumerate(self._parsed_active_data["data"])
        }
        DataJson()[self.widget_id]["total"] = self._rows_total
        self._maybe_update_selected_row()
        StateJson().send_changes()

    def _prepare_json_data(self, data: dict, key: str):
        if key in ("data", "columns"):
            default_value = []
        else:
            default_value = None
        source_data = data.get(key, default_value)
        if key == "data":
            source_data = self._sort_table_data(
                pd.DataFrame(data=source_data, columns=self._multi_idx_columns)
            )
        if key == "options":
            options = data.get(key, default_value)
            if options is not None:
                sort = options.get("sort", None)
                if sort is not None:
                    column_idx = sort.get("columnIndex", None)
                    if column_idx is not None:
                        sort["column"] = sort.get("columnIndex")
                        sort.pop("columnIndex")
        return source_data

    def _validate_sort(
        self, column_idx: int = None, order: Optional[Literal["asc", "desc"]] = None
    ):
        if column_idx is not None and type(column_idx) is not int:
            logger.warning(
                f"Incorrect value of 'column_id': {type(column_idx)} is not 'int'. Set to 'None'"
            )
            return False
        if column_idx is not None and column_idx < 0:
            logger.warning(f"Incorrect value of 'column_id': {column_idx} < 0. Set to 'None'")
            return False
        if order is not None and order not in ["asc", "desc"]:
            logger.warning(
                f"Incorrect value of 'order': {order}. Value can be one of 'asc' or 'desc'. Set to 'None'"
            )
            return False
        return True

    def _validate_sort_attrs(self):
        if not self._validate_sort(column_idx=self._sort_column_idx):
            self._sort_column_idx = None
        if not self._validate_sort(order=self._sort_order):
            self._sort_order = None

    def _validate_input_data(self, data):
        input_data_type = type(data)

        if input_data_type not in self._supported_types:
            raise TypeError(
                f"Cannot parse input data, please use one of supported datatypes: {self._supported_types}\n"
                """
                            1. Pandas Dataframe \n
                            2. Python list with structure [
                                                            ['row_1_column_1', 'row_1_column_2', ...],
                                                            ['row_2_column_1', 'row_2_column_2', ...],
                                                            ...
                                                          ]
                            """
            )

    def _create_multi_idx_columns(self):
        if self._columns_first_idx is not None:
            self._columns_second_idx = list(range(len(self._columns_first_idx)))
            _columns_tuples = [self._columns_first_idx, self._columns_second_idx]
            multi_idx_columns = pd.MultiIndex.from_tuples(
                list(zip(*_columns_tuples)), names=["first", "second"]
            )
        else:
            multi_idx_columns = None
        return multi_idx_columns

    def _prepare_input_data(self, data):
        if isinstance(data, pd.DataFrame):
            source_data = copy.deepcopy(data)

            if self._columns_first_idx is None and isinstance(source_data.columns, pd.RangeIndex):
                self._columns_first_idx = None
                self._multi_idx_columns = None
                return source_data
            elif self._columns_first_idx is None and isinstance(source_data.columns, pd.Index):
                self._columns_first_idx = source_data.columns.tolist()
            self._multi_idx_columns = self._create_multi_idx_columns()
            source_data.columns = self._multi_idx_columns
        elif isinstance(data, (list, type(None))):
            self._multi_idx_columns = self._create_multi_idx_columns()
            source_data = self._sort_table_data(
                pd.DataFrame(data=data, columns=self._multi_idx_columns)
            )
        return source_data

    def _prepare_working_data(self):
        parsed_source_data = self._unpack_pandas_table_data(input_data=self._source_data)
        sliced_data = self._slice_table_data(self._source_data, self._active_page)
        parsed_active_data = self._unpack_pandas_table_data(sliced_data)
        return parsed_source_data, sliced_data, parsed_active_data

    def _get_pandas_unpacked_data(self, data: pd.DataFrame) -> dict:
        if not isinstance(data, pd.DataFrame):
            raise TypeError("Cannot parse input data, please use Pandas Dataframe as input data")
        data = data.replace({np.nan: None})
        # data = data.astype(object).replace(np.nan, "-") # TODO: replace None later

        unpacked_data = {
            "columns": data.columns.to_list(),
            "data": data.values.tolist(),
        }
        return unpacked_data

    def _unpack_pandas_table_data(self, input_data: pd.DataFrame) -> dict:
        """
        Convert data to dict from pd.DataFrame

        """
        if len(input_data.columns) != 0:
            new_data = []
            row_idx_list = input_data.index.tolist()
            try:
                columns = input_data.columns.get_level_values("first").tolist()
            except KeyError:
                columns = []
            data = copy.deepcopy(self._get_pandas_unpacked_data(input_data))
            for idx, row in zip(row_idx_list, data["data"]):
                data_entity = {"idx": idx, "items": row}
                new_data.append(data_entity)
            data["data"] = new_data
            data["columns"] = columns
        else:
            data = {
                "columns": [],
                "data": [],
            }
        return data

    def _slice_table_data(self, input_data: pd.DataFrame, actual_page: int = 1) -> pd.DataFrame:
        """
        Prepare data rows for the active page according to all filters that have been set

        """
        if input_data is not None:
            data: pd.DataFrame = copy.deepcopy(input_data)
            end_idx = self._page_size * actual_page
            if end_idx == self._page_size:
                start_idx = 0
            else:
                start_idx = end_idx - self._page_size
            data = data.iloc[start_idx:end_idx]
        return data

    def _sort_table_data(
        self,
        input_data: pd.DataFrame,
        column_index: Optional[int] = None,
        sort_order: Optional[Literal["asc", "desc"]] = None,
    ) -> pd.DataFrame:
        """
        Apply sorting to received data

        """
        if column_index is None:
            column_index = self._sort_column_idx
        if sort_order is None:
            sort_order = self._sort_order

        if sort_order is None or column_index is None:
            return input_data  # unsorted

        data = copy.deepcopy(input_data)
        if input_data is None:
            return data

        data = self._sort_function(data=input_data, column_idx=column_index, order=sort_order)
        return data

    def _unpack_project_meta(self, project_meta: Union[ProjectMeta, dict]) -> dict:
        """
        Apply sorting to received data

        """
        if project_meta is not None:
            if isinstance(project_meta, ProjectMeta):
                project_meta = project_meta.to_json()
            elif not isinstance(project_meta, dict):
                raise TypeError(
                    f"Project meta possible types: dict, ProjectMeta. Instead got {type(project_meta)}"
                )
        return project_meta

    def _validate_fix_columns_value(self, fixed_columns: int) -> int:
        if fixed_columns is None:
            return None
        if isinstance(fixed_columns, int):
            if fixed_columns <= 0:
                logger.warning(
                    f"The value for 'fixed_columns' should be > 0, instead '{fixed_columns}' is obtained. Set the value to 'None'"
                )
                fixed_columns = None
            elif fixed_columns > 1:
                logger.warning(
                    f"Value for 'fixed_columns' is currently only supported as '1', instead '{fixed_columns}' is obtained. Set the value to '1'"
                )
                fixed_columns = 1
        else:
            logger.warning(
                f"The value for 'fixed_columns' should be 'int', instead '{type(fixed_columns)}' is obtained. Set the value to 'None'"
            )
            fixed_columns = None
        return fixed_columns

    def _validate_table_sizes(self, row):
        if len(row) != len(self._source_data.columns):
            raise ValueError(
                "Sizes mismatch:\n"
                f"Lenght of row -> {len(row)} != {len(self._source_data.columns)} <- leght of columns"
            )

    def _validate_row_values_types(self, row):
        failed_column_idxs = []
        failed_column_idx = 0
        for column, value in zip(self._source_data.columns, row):
            if len(self._source_data[column].values) == 0:
                continue
            col_type = type(self._source_data[column].values[0])
            if col_type == str and not isinstance(value, str):
                failed_column_idxs.append(
                    {
                        "idx": failed_column_idx,
                        "type": col_type.__name__,
                        "row_value_type": type(value).__name__,
                    }
                )
            elif col_type != str and isinstance(value, str):
                failed_column_idxs.append(
                    {
                        "idx": failed_column_idx,
                        "type": col_type.__name__,
                        "row_value_type": type(value).__name__,
                    }
                )
            failed_column_idx += 1
        if len(failed_column_idxs) != 0:
            raise TypeError(
                f"Row contains values of types that do not match the data types in the columns: {failed_column_idxs}"
            )

    def update_cell_value(self, row: int, column: int, value: Any) -> None:
        """Updates the value of the cell in the table.

        :param row: Index of the row
        :type row: int
        :param column: Index of the column
        :type column: int
        :param value: New value
        :type value: Any
        """

        self._source_data.iat[row, column] = value
        self._parsed_source_data = self._unpack_pandas_table_data(self._source_data)
        self._sort_column_idx = StateJson()[self.widget_id]["sort"]["column"]
        self._sort_order = StateJson()[self.widget_id]["sort"]["order"]
        self._validate_sort_attrs()
        self._filtered_data = self._filter(self._filter_value)
        self._searched_data = self._search(self._search_str)
        self._rows_total = len(self._searched_data)
        self._sorted_data = self._sort_table_data(self._searched_data)

        increment = 0 if self._rows_total % self._page_size == 0 else 1
        max_page = self._rows_total // self._page_size + increment
        if (
            self._active_page > max_page
        ):  # active page is out of range (in case of the filtered data)
            self._active_page = max_page
            StateJson()[self.widget_id]["page"] = self._active_page

        self._sliced_data = self._slice_table_data(self._sorted_data, actual_page=self._active_page)
        self._parsed_active_data = self._unpack_pandas_table_data(self._sliced_data)
        DataJson()[self.widget_id]["data"] = {
            i: row for i, row in enumerate(self._parsed_active_data["data"])
        }
        DataJson()[self.widget_id]["total"] = self._rows_total
        DataJson().send_changes()
        StateJson().send_changes()

    def selection_changed(self, func):
        """Decorator for function that handles selection change event.

        :param func: Function that handles selection change event
        :type func: Callable[[], Any]
        :return: Decorated function
        :rtype: Callable[[], None]
        """
        selection_changed_route_path = self.get_route_path(FastTable.Routes.SELECTION_CHANGED)
        server = self._sly_app.get_server()

        @server.post(selection_changed_route_path)
        def _selection_changed():
            if self._is_radio:
                selected_row = self.get_selected_row()
                func(selected_row)
            elif self._is_selectable:
                selected_rows = self.get_selected_rows()
                func(selected_rows)

        self._selection_changed_handled = True
        DataJson()[self.widget_id]["selectionChangedHandled"] = True
        DataJson().send_changes()
        return _selection_changed

    def select_row(self, idx: int):
        if not self._is_selectable and not self._is_radio:
            raise ValueError(
                "Table is not selectable. Set 'is_selectable' or 'is_radio' to True to use this method."
            )
        if idx < 0 or idx >= len(self._parsed_source_data["data"]):
            raise IndexError(
                f"Row index {idx} is out of range. Valid range is 0 to {len(self._parsed_source_data['data']) - 1}."
            )
        selected_row = self._parsed_source_data["data"][idx]
        self._selected_rows = [
            {"idx": idx, "row": selected_row.get("items", selected_row.get("row", None))}
        ]
        StateJson()[self.widget_id]["selectedRows"] = self._selected_rows
        page = idx // self._page_size + 1
        if self._active_page != page:
            self._active_page = page
            StateJson()[self.widget_id]["page"] = self._active_page
        self._refresh()

    def select_rows(self, idxs: List[int]):
        if not self._is_selectable:
            raise ValueError(
                "Table is not selectable. Set 'is_selectable' to True to use this method."
            )
        selected_rows = [
            self._parsed_source_data["data"][idx]
            for idx in idxs
            if 0 <= idx < len(self._parsed_source_data["data"])
        ]
        self._selected_rows = [
            {"idx": row["idx"], "row": row.get("items", row.get("row", None))}
            for row in selected_rows
        ]
        StateJson()[self.widget_id]["selectedRows"] = self._selected_rows
        StateJson().send_changes()

    def select_row_by_value(self, column, value: Any):
        """Selects a row by value in a specific column.

        :param column: Column name to filter by
        :type column: str
        :param value: Value to select row by
        :type value: Any
        """
        if not self._is_selectable and not self._is_radio:
            raise ValueError(
                "Table is not selectable. Set 'is_selectable' to True to use this method."
            )
        if column not in self._columns_first_idx:
            raise ValueError(f"Column '{column}' does not exist in the table.")

        idx = self._source_data[self._source_data[column] == value].index.tolist()
        if not idx:
            raise ValueError(f"No rows found with {column} = {value}.")
        if len(idx) > 1:
            raise ValueError(
                f"Multiple rows found with {column} = {value}. Please use select_rows_by_value method."
            )
        self.select_row(idx[0])

    def select_rows_by_value(self, column, values: List):
        """Selects rows by value in a specific column.

        :param column: Column name to filter by
        :type column: str
        :param values: List of values to select rows by
        :type values: List
        """
        if not self._is_selectable:
            raise ValueError(
                "Table is not selectable. Set 'is_selectable' to True to use this method."
            )
        if column not in self._columns_first_idx:
            raise ValueError(f"Column '{column}' does not exist in the table.")

        idxs = self._source_data[self._source_data[column].isin(values)].index.tolist()
        self.select_rows(idxs)

    def _read_custom_columns(self, columns: List[Union[str, tuple]]) -> None:
        if not columns:
            return
        self._columns = columns
        self._columns_options = self._columns_options or [{} for _ in columns]
        self._columns_data = []
        self._columns_first_idx = []
        for i, col in enumerate(columns):
            if isinstance(col, str):
                self._columns_first_idx.append(col)
                self._columns_data.append(self.ColumnData(name=col))
            elif isinstance(col, tuple):
                self._columns_first_idx.append(col[0])
                self._columns_data.append(
                    self.ColumnData(name=col[0], is_widget=True, widget=col[1])
                )
                self._columns_options[i]["customCell"] = True
            else:
                raise TypeError(f"Column name must be a string or a tuple, got {type(col)}")

        self._validate_sort_attrs()



# coding: utf-8
"""api connection to the server which allows user to communicate with Supervisely"""

from __future__ import annotations

import asyncio
import datetime
import gc
import glob
import json
import os
import shutil
import threading
from logging import Logger
from pathlib import Path
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterable,
    Dict,
    Generator,
    Iterable,
    Literal,
    Mapping,
    Optional,
    Union,
)
from urllib.parse import urljoin, urlparse

import httpx
import jwt
import requests
from dotenv import get_key, load_dotenv, set_key
from pkg_resources import parse_version
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

import supervisely.api.advanced_api as advanced_api
import supervisely.api.agent_api as agent_api
import supervisely.api.annotation_api as annotation_api
import supervisely.api.app_api as app_api
import supervisely.api.dataset_api as dataset_api
import supervisely.api.entities_collection_api as entities_collection_api
import supervisely.api.file_api as file_api
import supervisely.api.github_api as github_api
import supervisely.api.image_annotation_tool_api as image_annotation_tool_api
import supervisely.api.image_api as image_api
import supervisely.api.import_storage_api as import_stoarge_api
import supervisely.api.issues_api as issues_api
import supervisely.api.labeling_job_api as labeling_job_api
import supervisely.api.labeling_queue_api as labeling_queue_api
import supervisely.api.nn.neural_network_api as neural_network_api
import supervisely.api.object_class_api as object_class_api
import supervisely.api.plugin_api as plugin_api
import supervisely.api.pointcloud.pointcloud_api as pointcloud_api
import supervisely.api.pointcloud.pointcloud_episode_api as pointcloud_episode_api
import supervisely.api.project_api as project_api
import supervisely.api.remote_storage_api as remote_storage_api
import supervisely.api.report_api as report_api
import supervisely.api.role_api as role_api
import supervisely.api.storage_api as storage_api
import supervisely.api.task_api as task_api
import supervisely.api.team_api as team_api
import supervisely.api.user_api as user_api
import supervisely.api.video.video_api as video_api
import supervisely.api.video_annotation_tool_api as video_annotation_tool_api
import supervisely.api.volume.volume_api as volume_api
import supervisely.api.workspace_api as workspace_api
import supervisely.io.env as sly_env
from supervisely._utils import camel_to_snake, is_community, is_development
from supervisely.api.module_api import ApiField
from supervisely.io.network_exceptions import (
    RetryableRequestException,
    process_requests_exception,
    process_requests_exception_async,
    process_unhandled_request,
)
from supervisely.project.project_meta import ProjectMeta
from supervisely.sly_logger import logger

SUPERVISELY_TASK_ID = "SUPERVISELY_TASK_ID"
SUPERVISELY_PUBLIC_API_RETRIES = "SUPERVISELY_PUBLIC_API_RETRIES"
SUPERVISELY_PUBLIC_API_RETRY_SLEEP_SEC = "SUPERVISELY_PUBLIC_API_RETRY_SLEEP_SEC"
SERVER_ADDRESS = "SERVER_ADDRESS"
SUPERVISELY_API_SERVER_ADDRESS = "SUPERVISELY_API_SERVER_ADDRESS"
API_TOKEN = "API_TOKEN"
TASK_ID = "TASK_ID"
SUPERVISELY_ENV_FILE = os.path.join(Path.home(), "supervisely.env")


class ApiContext:
    """
    Context manager for the API object for optimization purposes.
    Use this context manager when you need to perform a series of operations on the same project or dataset.
    It allows you to avoid redundant API calls to get the same project or dataset info multiple times.

    :param api: API object.
    :type api: :class:`Api`
    :param project_id: Project ID.
    :type project_id: int, optional
    :param dataset_id: Dataset ID.
    :type dataset_id: int, optional
    :param project_meta: ProjectMeta object.
    :type project_meta: :class:`ProjectMeta`, optional
    :raises: :class:`RuntimeError`, if api is None.

    :Usage example:

         .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            if sly.is_development():
                load_dotenv(os.path.expanduser("~/supervisely.env"))
            api = sly.Api.from_env()

            with ApiContext(
                api,
                project_id=33333,
                dataset_id=99999,
                project_meta=project_meta,
                with_alpha_masks=True,
            ):
                api.annotation.upload_paths(image_ids, ann_paths, anns_progress)
                # another code here
    """

    def __init__(
        self,
        api: Api,
        project_id: Optional[int] = None,
        dataset_id: Optional[int] = None,
        project_meta: Optional[ProjectMeta] = None,
        with_alpha_masks: Optional[bool] = True,
    ):
        if api is None:
            raise RuntimeError("Api object is None")
        self.api = api
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.project_meta = project_meta
        self.with_alpha_masks = with_alpha_masks

    def __enter__(self):
        self.api.optimization_context = {
            "project_id": self.project_id,
            "dataset_id": self.dataset_id,
            "project_meta": self.project_meta,
            "with_alpha_masks": self.with_alpha_masks,
        }
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.api.optimization_context = {}


class UserSession:
    """
    UserSession object contains info that is returned after user authentication.

    :param server: Server url.
    :type server: str
    :raises: :class:`RuntimeError`, if server url is invalid.
    """

    def __init__(self, server_address: str):
        self.api_token = None
        self.team_id = None
        self.workspace_id = None
        self.server_address = server_address

        if not self._normalize_and_validate_server_url():
            raise RuntimeError(f"Invalid server url: {server_address}")

    def __str__(self):
        return f"UserSession(server={self.server_address})"

    def __repr__(self):
        return self.__str__()

    def _normalize_and_validate_server_url(self) -> bool:
        """
        Validate server url.

        :return: True if server url is valid, False otherwise.
        """
        self.server_address = Api.normalize_server_address(self.server_address)
        if not self.server_address.startswith("https://"):
            response = requests.get(self.server_address, allow_redirects=False)
            if (300 <= response.status_code < 400) or (
                response.headers.get("Location", "").startswith("https://")
            ):
                self.server_address = self.server_address.replace("http://", "https://")
        result = urlparse(self.server_address)
        if all([result.scheme, result.netloc]):
            try:
                response = requests.get(self.server_address)
                if response.status_code == 200:
                    return True
            except requests.RequestException:
                pass
        return False

    def _setattrs_user_session(self, decoded_token):
        """
        Add decoded info to UserSession object.

        :param decoded_token: Decoded token.
        :type decoded_token: dict
        :return: None
        :rtype: :class:`NoneType`
        """
        for key, value in decoded_token.items():
            if key == "group":
                self.team = value
                self.team_id = value["id"]
            elif key == "workspace":
                self.workspace = value
                self.workspace_id = value["id"]
            else:
                key = camel_to_snake(key)
                setattr(self, key, value)

    def log_in(self, login: str, password: str) -> UserSession:
        """
        Authenticate user and return UserSession object with decoded info from JWT token.

        :param login: User login.
        :type login: str
        :param password: User password.
        :type password: str
        :return: UserSession object
        :rtype: :class:`UserSession`
        """
        login_url = urljoin(self.server_address, "api/account")
        payload = {"login": login, "password": password}
        response = requests.post(login_url, data=payload)
        del password
        gc.collect()
        if response.status_code == 200:
            data = response.json()
            jwt_token = data.get("token", None)
            decoded_token = jwt.decode(jwt_token, options={"verify_signature": False})
            self._setattrs_user_session(decoded_token)
            return self
        else:
            raise RuntimeError(f"Failed to authenticate user: status code {response.status_code}")


class Api:
    """
    An API connection to the server with which you can communicate with your teams, workspaces and projects. :class:`Api<Api>` object is immutable.

    :param server_address: Address of the server.
    :type server_address: str
    :param token: Unique secret token associated with your agent.
    :type token: str
    :param retry_count: The number of attempts to connect to the server.
    :type retry_count: int, optional
    :param retry_sleep_sec: The number of seconds to delay between attempts to connect to the server.
    :type retry_sleep_sec: int, optional
    :param external_logger: Logger class object.
    :type external_logger: logger, optional
    :param ignore_task_id:
    :type ignore_task_id: bool, optional
    :param api_server_address: Address of the API server.
    :type api_server_address: str, optional
    :param check_instance_version: Check if the given version is lower or equal to the current
        Supervisely instance version. If set to True, will try to read the version from the environment variable
        "MINIMUM_INSTANCE_VERSION_FOR_SDK". If set to a string, will use this string as the version to check.
        If set to False, will skip the check.
    :type check_instance_version: bool or str, optional
    :raises: :class:`ValueError`, if token is None or it length != 128
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervisely.com", token="4r47N...xaTatb")
    """

    _checked_servers = set()

    def __init__(
        self,
        server_address: Optional[str] = None,
        token: Optional[str] = None,
        retry_count: Optional[int] = 10,
        retry_sleep_sec: Optional[int] = None,
        external_logger: Optional[Logger] = None,
        ignore_task_id: bool = False,
        api_server_address: Optional[str] = None,
        check_instance_version: Union[bool, str] = False,
    ):
        self.logger = external_logger or logger

        if server_address is None:
            server_address = os.environ.get(SERVER_ADDRESS, None)
        if token is None:
            token = os.environ.get(API_TOKEN, None)

        if server_address is None:
            raise ValueError(
                "SERVER_ADDRESS env variable is undefined, https://developer.supervisely.com/getting-started/basics-of-authentication"
            )

        self.server_address = Api.normalize_server_address(server_address)

        self._api_server_address = None
        if api_server_address is None:
            api_server_address = os.environ.get(SUPERVISELY_API_SERVER_ADDRESS, None)

        if api_server_address is not None:
            self._api_server_address = Api.normalize_server_address(api_server_address)

        if retry_count is None:
            retry_count = int(os.getenv(SUPERVISELY_PUBLIC_API_RETRIES, "10"))
        if retry_sleep_sec is None:
            retry_sleep_sec = int(os.getenv(SUPERVISELY_PUBLIC_API_RETRY_SLEEP_SEC, "1"))

        self.token = token
        self.headers = {}
        if token is not None:
            self.headers["x-api-key"] = token
        self.task_id = os.getenv(SUPERVISELY_TASK_ID)
        if self.task_id is not None and ignore_task_id is False:
            self.headers["x-task-id"] = self.task_id
        self.context = {}
        self.additional_fields = {}
        self.optimization_context = {}

        self.team = team_api.TeamApi(self)
        self.workspace = workspace_api.WorkspaceApi(self)
        self.project = project_api.ProjectApi(self)
        self.nn = neural_network_api.NeuralNetworkApi(self)
        self.task = task_api.TaskApi(self)
        self.dataset = dataset_api.DatasetApi(self)
        self.image = image_api.ImageApi(self)
        self.annotation = annotation_api.AnnotationApi(self)
        self.plugin = plugin_api.PluginApi(self)
        self.agent = agent_api.AgentApi(self)
        self.role = role_api.RoleApi(self)
        self.user = user_api.UserApi(self)
        self.labeling_job = labeling_job_api.LabelingJobApi(self)
        self.labeling_queue = labeling_queue_api.LabelingQueueApi(self)
        self.video = video_api.VideoApi(self)
        # self.project_class = project_class_api.ProjectClassApi(self)
        self.object_class = object_class_api.ObjectClassApi(self)
        self.report = report_api.ReportApi(self)
        self.pointcloud = pointcloud_api.PointcloudApi(self)
        self.pointcloud_episode = pointcloud_episode_api.PointcloudEpisodeApi(self)
        self.app = app_api.AppApi(self)
        self.file = file_api.FileApi(self)
        self.storage = storage_api.StorageApi(self)
        self.img_ann_tool = image_annotation_tool_api.ImageAnnotationToolApi(self)
        self.vid_ann_tool = video_annotation_tool_api.VideoAnnotationToolApi(self)
        self.advanced = advanced_api.AdvancedApi(self)
        self.import_storage = import_stoarge_api.ImportStorageApi(self)
        self.remote_storage = remote_storage_api.RemoteStorageApi(self)
        self.github = github_api.GithubApi(self)
        self.volume = volume_api.VolumeApi(self)
        self.issues = issues_api.IssuesApi(self)
        self.entities_collection = entities_collection_api.EntitiesCollectionApi(self)

        self.retry_count = retry_count
        self.retry_sleep_sec = retry_sleep_sec

        skip_from_env = sly_env.supervisely_skip_https_user_helper_check()
        self._skip_https_redirect_check = (
            skip_from_env or self.server_address in Api._checked_servers
        )
        self.logger.trace(
            f"Skip HTTPS redirect check on API init: {self._skip_https_redirect_check}. ENV: {skip_from_env}. Checked servers: {Api._checked_servers}"
        )
        self._require_https_redirect_check = (
            False
            if self._skip_https_redirect_check
            else not self.server_address.startswith("https://")
        )

        self.async_httpx_client: httpx.AsyncClient = None
        self.httpx_client: httpx.Client = None
        self._semaphore = None
        self._instance_version = None
        self._version_check_completed = False
        self._version_check_lock = threading.Lock()

        if check_instance_version:
            self._check_version(None if check_instance_version is True else check_instance_version)

    @classmethod
    def normalize_server_address(cls, server_address: str) -> str:
        """ """
        result = server_address.strip("/")
        if ("http://" not in result) and ("https://" not in result):
            result = "http://" + result
        return result

    @classmethod
    def from_env(
        cls,
        retry_count: int = 10,
        ignore_task_id: bool = False,
        env_file: str = SUPERVISELY_ENV_FILE,
        check_instance_version: Union[bool, str] = False,
    ) -> Api:
        """
        Initialize API use environment variables.

        :param retry_count: The number of attempts to connect to the server.
        :type retry_count: int
        :param ignore_task_id:
        :type ignore_task_id: bool
        :param env_file: Path to your .env file.
        :type env_file: str
        :param check_instance_version: Check if the given version is lower or equal to the current
            version of the Supervisely instance.
        :type check_instance_version: bool or str, optional
        :return: Api object
        :rtype: :class:`Api<supervisely.api.api.Api>`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'

            api = sly.Api.from_env()

            # alternatively you can store SERVER_ADDRESS and API_TOKEN
            # in "~/supervisely.env" .env file
            # Learn more here: https://developer.supervisely.com/app-development/basics/add-private-app#create-.env-file-supervisely.env-with-the-following-content-learn-more-here

            api = sly.Api.from_env()
        """

        server_address = sly_env.server_address(raise_not_found=False)
        token = sly_env.api_token(raise_not_found=False)

        if is_development() and None in (server_address, token):
            env_path = os.path.expanduser(env_file)
            if os.path.exists(env_path):
                _, extension = os.path.splitext(env_path)
                if extension == ".env":
                    load_dotenv(env_path)
                    server_address = sly_env.server_address()
                    token = sly_env.api_token()
                else:
                    raise ValueError(f"'{env_path}' is not an '*.env' file")
            else:
                raise FileNotFoundError(f"File not found: '{env_path}'")

        if server_address is None:
            raise ValueError(
                "SERVER_ADDRESS env variable is undefined. Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication"
            )
        if token is None:
            raise ValueError(
                "API_TOKEN env variable is undefined. Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication"
            )

        return cls(
            server_address,
            token,
            retry_count=retry_count,
            ignore_task_id=ignore_task_id,
            check_instance_version=check_instance_version,
        )

    def add_header(self, key: str, value: str) -> None:
        """
        Add given key and value to headers dictionary.

        :param key: New key.
        :type key: str
        :param value: New value.
        :type value: str
        :raises: :class:`RuntimeError`, if key is already set
        :return: None
        :rtype: :class:`NoneType`
        """
        if key in self.headers:
            raise RuntimeError(
                f"Header {key!r} is already set for the API object. "
                f"Current value: {self.headers[key]!r}. Tried to set value: {value!r}"
            )
        self.headers[key] = value

    def add_additional_field(self, key: str, value: str) -> None:
        """
        Add given key and value to additional_fields dictionary.

        :param key: New key.
        :type key: str
        :param value: New value.
        :type value: str
        :return: None
        :rtype: :class:`NoneType`
        """
        self.additional_fields[key] = value

    @property
    def instance_version(self) -> str:
        """Return Supervisely instance version, e.g. "6.9.13".
        If the version cannot be determined, return "unknown".

        :return: Supervisely instance version or "unknown" if the version cannot be determined.
        :rtype: str

        :Usage example:

        .. code-block:: python

                import supervisely as sly

                api = sly.Api(server_address='https://app.supervisely.com', token='4r47N...xaTatb')
                print(api.instance_version)
                # Output:
                # '6.9.13'
        """
        try:
            if self._instance_version is None:
                self._instance_version = (
                    self.post("instance.version", {}).json().get(ApiField.VERSION)
                )
        except Exception as e:
            logger.warning(f"Failed to get instance version from server: {e}")
            self._instance_version = "unknown"
        return self._instance_version

    def is_version_supported(self, version: Optional[str] = None) -> Union[bool, None]:
        """Check if the given version is lower or equal to the current Supervisely instance version.
        If the version omitted, will try to read it from the environment variable "MINIMUM_INSTANCE_VERSION_FOR_SDK".
        If the version is lower or equal, return True, otherwise False.
        If the version of the instance cannot be determined, return False.

        :param version: Version to check.
        :type version: Optional[str], e.g. "6.9.13"
        :return: True if the given version is lower or equal to the current Supervisely
            instance version, otherwise False.
        :rtype: bool

        :Usage example:

        .. code-block:: python

            import supervisely as sly

            api = sly.Api(server_address='https://app.supervisely.com', token='4r47N...xaTatb')
            version_to_check = "6.9.13"
            print(api.is_version_supported(version_to_check))
            # Output:
            # True
        """
        instance_version = self.instance_version
        if instance_version == "unknown":
            return

        if not version:
            version = sly_env.mininum_instance_version_for_sdk()
            if not version:
                logger.debug(
                    "Cant find MINIMUM_INSTANCE_VERSION_FOR_SDK in environment variables, "
                    "check of the minimum version is skipped."
                )
                return

        try:
            version = str(version)
        except Exception:
            logger.warning(
                f"Provided version {version!r} is not a valid version string "
                f"(expected format: 'x.y.z'). The output of this function will be incorrect."
            )
            return

        return parse_version(instance_version) >= parse_version(version)

    def _check_version(self, version: Optional[str] = None) -> None:
        """Check if the given version is compatible with the current Supervisely instance version.
        Compatible means that the given version is lower or equal to the current Supervisely instance version.
        If check was not successful, log a debug message, if the version is not supported, log a warning message.

        :param version: Version to check.
        :type version: Optional[str], e.g. "6.9.13"
        """

        # Thread-safe one-time check with double-checked locking pattern
        if self._version_check_completed:
            return

        with self._version_check_lock:
            # Double-check inside the lock
            if self._version_check_completed:
                return

            self._version_check_completed = True

            # Since it's a informational message, we don't raise an exception if the check fails
            # in any case, we don't want to interrupt the user's workflow.
            try:
                check_result = self.is_version_supported(version)
                if check_result is None:
                    logger.debug(
                        "Failed to check if the instance version meets the minimum requirements "
                        "of current SDK version. "
                        "Ensure that the MINIMUM_INSTANCE_VERSION_FOR_SDK environment variable is set. "
                        "Usually you can ignore this message, but if you're adding new features, "
                        "which will require upgrade of the Supervisely instance, you should update "
                        "it supervisely.__init__.py file."
                    )
                if check_result is False:
                    message = (
                        "The current version of the Supervisely instance is not supported by the SDK. "
                        "Some features may not work correctly."
                    )
                    if not is_community():
                        message += (
                            " Please upgrade the Supervisely instance to the latest version (recommended) "
                            "or downgrade the SDK to the version that supports the current instance (not recommended). "
                            "Refer to this docs for more information: "
                            "https://docs.supervisely.com/enterprise-edition/get-supervisely/upgrade "
                            "Check out changelog for the latest version of Supervisely: "
                            "https://app.supervisely.com/changelog"
                        )
                        logger.warning(message)
            except Exception as e:
                logger.debug(
                    f"Tried to check version compatibility between SDK and instance, but failed: {e}"
                )

    def post(
        self,
        method: str,
        data: Dict,
        retries: Optional[int] = None,
        stream: Optional[bool] = False,
        raise_error: Optional[bool] = False,
    ) -> requests.Response:
        """
        Performs POST request to server with given parameters.

        :param method: Method name.
        :type method: str
        :param data: Dictionary to send in the body of the :class:`Request`.
        :type data: dict
        :param retries: The number of attempts to connect to the server.
        :type retries: int, optional
        :param stream: Define, if you'd like to get the raw socket response from the server.
        :type stream: bool, optional
        :param raise_error: Define, if you'd like to raise error if connection is failed. Retries will be ignored.
        :type raise_error: bool, optional
        :return: Response object
        :rtype: :class:`Response<Response>`
        """
        if not self._skip_https_redirect_check:
            self._check_https_redirect()
        if retries is None:
            retries = self.retry_count

        url = self.api_server_address + "/v3/" + method
        logger.trace(f"POST {url}")

        for retry_idx in range(retries):
            response = None
            try:
                if type(data) is bytes:
                    response = requests.post(url, data=data, headers=self.headers, stream=stream)
                elif type(data) is MultipartEncoderMonitor or type(data) is MultipartEncoder:
                    response = requests.post(
                        url,
                        data=data,
                        headers={**self.headers, "Content-Type": data.content_type},
                        stream=stream,
                    )
                else:
                    json_body = data
                    if type(data) is dict:
                        json_body = {**data, **self.additional_fields}
                    response = requests.post(
                        url, json=json_body, headers=self.headers, stream=stream
                    )

                if response.status_code != requests.codes.ok:  # pylint: disable=no-member
                    if not self._version_check_completed:
                        self._check_version()
                    Api._raise_for_status(response)
                return response
            except requests.RequestException as exc:
                if (
                    isinstance(exc, requests.exceptions.HTTPError)
                    and response.status_code == 400
                    and self.token is None
                ):
                    self.logger.warning(
                        "API_TOKEN env variable is undefined. See more: https://developer.supervisely.com/getting-started/basics-of-authentication"
                    )
                if raise_error:
                    raise exc
                else:
                    process_requests_exception(
                        self.logger,
                        exc,
                        method,
                        url,
                        verbose=True,
                        swallow_exc=True,
                        sleep_sec=min(self.retry_sleep_sec * (2**retry_idx), 60),
                        response=response,
                        retry_info={"retry_idx": retry_idx + 1, "retry_limit": retries},
                    )
            except Exception as exc:
                process_unhandled_request(self.logger, exc)
        raise requests.exceptions.RetryError("Retry limit exceeded ({!r})".format(url))

    def get(
        self,
        method: str,
        params: Dict,
        retries: Optional[int] = None,
        stream: Optional[bool] = False,
        use_public_api: Optional[bool] = True,
        data: Optional[Dict] = None,
    ) -> requests.Response:
        """
        Performs GET request to server with given parameters.

        :param method:
        :type method: str
        :param params: Dictionary to send in the body of the :class:`Request`.
        :type params: dict
        :param retries: The number of attempts to connect to the server.
        :type retries: int, optional
        :param stream: Define, if you'd like to get the raw socket response from the server.
        :type stream: bool, optional
        :param use_public_api:
        :type use_public_api: bool, optional
        :param data: Dictionary to send in the body of the :class:`Request`.
        :type data: dict, optional
        :return: Response object
        :rtype: :class:`Response<Response>`
        """
        if not self._skip_https_redirect_check:
            self._check_https_redirect()
        if retries is None:
            retries = self.retry_count

        url = self.api_server_address + "/v3/" + method
        if use_public_api is False:
            url = os.path.join(self.server_address, method)
        logger.trace(f"GET {url}")

        for retry_idx in range(retries):
            response = None
            try:
                json_body = params
                if type(params) is dict:
                    json_body = {**params, **self.additional_fields}
                response = requests.get(
                    url, params=json_body, data=data, headers=self.headers, stream=stream
                )

                if response.status_code != requests.codes.ok:  # pylint: disable=no-member
                    Api._raise_for_status(response)
                return response
            except requests.RequestException as exc:
                if (
                    isinstance(exc, requests.exceptions.HTTPError)
                    and response.status_code == 400
                    and self.token is None
                ):
                    self.logger.warning(
                        "API_TOKEN env variable is undefined. See more: https://developer.supervisely.com/getting-started/basics-of-authentication"
                    )
                process_requests_exception(
                    self.logger,
                    exc,
                    method,
                    url,
                    verbose=True,
                    swallow_exc=True,
                    sleep_sec=min(self.retry_sleep_sec * (2**retry_idx), 60),
                    response=response,
                    retry_info={"retry_idx": retry_idx + 2, "retry_limit": retries},
                )
            except Exception as exc:
                process_unhandled_request(self.logger, exc)

    @staticmethod
    def _raise_for_status(response: requests.Response):
        """
        Raise error and show message with error code if given response can not connect to server.
        :param response: Request class object
        """
        http_error_msg = ""
        if isinstance(response.reason, bytes):
            try:
                reason = response.reason.decode("utf-8")
            except UnicodeDecodeError:
                reason = response.reason.decode("iso-8859-1")
        else:
            reason = response.reason

        if 400 <= response.status_code < 500:
            http_error_msg = "%s Client Error: %s for url: %s (%s)" % (
                response.status_code,
                reason,
                response.url,
                response.content.decode("utf-8"),
            )

        elif 500 <= response.status_code < 600:
            http_error_msg = "%s Server Error: %s for url: %s (%s)" % (
                response.status_code,
                reason,
                response.url,
                response.content.decode("utf-8"),
            )

        if http_error_msg:
            raise requests.exceptions.HTTPError(http_error_msg, response=response)

    @staticmethod
    def _raise_for_status_httpx(response: httpx.Response):
        """
        Raise error and show message with error code if given response can not connect to server.
        :param response: Response class object
        """
        http_error_msg = ""

        if hasattr(response, "reason_phrase"):
            reason = response.reason_phrase
        else:
            reason = "Can't get reason"

        def decode_response_content(response: httpx.Response):
            try:
                return response.content.decode("utf-8")
            except Exception as e:
                if hasattr(response, "is_stream_consumed"):
                    return f"Stream is consumed. {e}"
                else:
                    return f"Can't decode response content: {e}"

        if 400 <= response.status_code < 500:
            http_error_msg = "%s Client Error: %s for url: %s (%s)" % (
                response.status_code,
                reason,
                response.url,
                decode_response_content(response),
            )

        elif 500 <= response.status_code < 600:
            http_error_msg = "%s Server Error: %s for url: %s (%s)" % (
                response.status_code,
                reason,
                response.url,
                decode_response_content(response),
            )

        if http_error_msg:
            raise httpx.HTTPStatusError(
                message=http_error_msg, response=response, request=response.request
            )

    @staticmethod
    def parse_error(
        response: requests.Response,
        default_error: Optional[str] = "Error",
        default_message: Optional[str] = "please, contact administrator",
    ):
        """
        Processes error from response.

        :param response: Request object.
        :type method: Request
        :param default_error: Error description.
        :type method: str, optional
        :param default_message: Message to user.
        :type method: str, optional
        :return: Number of error and message about curren connection mistake
        :rtype: :class:`int`, :class:`str`
        """
        ERROR_FIELD = "error"
        MESSAGE_FIELD = "message"
        DETAILS_FIELD = "details"

        try:
            data_str = response.content.decode("utf-8")
            data = json.loads(data_str)
            error = data.get(ERROR_FIELD, default_error)
            details = data.get(DETAILS_FIELD, {})
            if type(details) is dict:
                message = details.get(MESSAGE_FIELD, default_message)
            else:
                message = details[0].get(MESSAGE_FIELD, default_message)

            return error, message
        except Exception:
            return "", ""

    def pop_header(self, key: str) -> str:
        """ """
        if key not in self.headers:
            raise KeyError(f"Header {key!r} not found")
        return self.headers.pop(key)

    def _check_https_redirect(self):
        """
        Check if HTTP server should be redirected to HTTPS.
        If the server has already been checked before (for any instance of this class),
        skip the check to avoid redundant network requests.
        """
        if self._require_https_redirect_check is True:
            if self.server_address in Api._checked_servers:
                self._require_https_redirect_check = False
                return

            try:
                response = requests.get(
                    self.server_address.replace("http://", "https://"),
                    allow_redirects=False,
                    timeout=(5, 15),
                )
                response.raise_for_status()
                self.server_address = self.server_address.replace("http://", "https://")
                msg = (
                    "You're using HTTP server address while the server requires HTTPS. "
                    "Supervisely automatically changed the server address to HTTPS for you. "
                    f"Consider updating your server address to {self.server_address}"
                )
                self.logger.warning(msg)
            except:
                pass
            finally:
                Api._checked_servers.add(self.server_address)
                self._require_https_redirect_check = False

    @classmethod
    def from_credentials(
        cls,
        server_address: str,
        login: str,
        password: str,
        override: bool = False,
        env_file: str = SUPERVISELY_ENV_FILE,
        check_instance_version: Union[bool, str] = False,
    ) -> Api:
        """
        Create Api object using credentials and optionally save them to ".env" file with overriding environment variables.
        If ".env" file already exists, backup will be created automatically.
        All backups will be stored in the same directory with postfix "_YYYYMMDDHHMMSS". You can have not more than 5 last backups.
        This method can be used also to update ".env" file.

        :param server_address: Supervisely server url.
        :type server_address: str
        :param login: User login.
        :type login: str
        :param password: User password.
        :type password: str
        :param override: If False, return Api object. If True, additionally create ".env" file or overwrite existing (backup file will be created automatically), and override environment variables.
        :type override: bool, optional
        :param env_file: Path to your .env file.
        :type env_file: str, optional
        :param check_instance_version: Check if the given version is lower or equal to the current
            version of the Supervisely instance.
        :type check_instance_version: bool or str, optional
        :return: Api object

        :Usage example:

             .. code-block:: python

                import supervisely as sly

                server_address = 'https://app.supervisely.com'
                login = 'user'
                password = 'pass'

                api = sly.Api.from_credentials(server_address, login, password)
        """

        session = UserSession(server_address).log_in(login, password)
        del password
        gc.collect()

        api = cls(
            session.server_address,
            session.api_token,
            ignore_task_id=True,
            check_instance_version=check_instance_version,
        )

        if override:
            if os.path.isfile(env_file):
                # create backup
                timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
                backup_file = f"{env_file}_{timestamp}"
                shutil.copy2(env_file, backup_file)
                if api.token != get_key(env_file, API_TOKEN):
                    # create new file
                    os.remove(env_file)
                    Path(env_file).touch()
                # remove old backups
                all_backups = sorted(glob.glob(f"{env_file}_" + "[0-9]" * 14))
                while len(all_backups) > 5:
                    os.remove(all_backups.pop(0))
            set_key(env_file, SERVER_ADDRESS, session.server_address)
            set_key(env_file, API_TOKEN, session.api_token)
            if session.team_id:
                set_key(env_file, "INIT_GROUP_ID", f"{session.team_id}")
            if session.workspace_id:
                set_key(env_file, "INIT_WORKSPACE_ID", f"{session.workspace_id}")
            load_dotenv(env_file, override=override)
        return api

    @property
    def api_server_address(self) -> str:
        """
        Get API server address.

        :return: API server address.
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            api = sly.Api(server_address='https://app.supervisely.com', token='4r47N...xaTatb')
            print(api.api_server_address)
            # Output:
            # 'https://app.supervisely.com/public/api'
        """

        if self._api_server_address is not None:
            return self._api_server_address

        return f"{self.server_address}/public/api"

    def post_httpx(
        self,
        method: str,
        json: Dict = None,
        content: Union[str, bytes, Iterable[bytes], AsyncIterable[bytes]] = None,
        files: Union[Mapping] = None,
        params: Union[str, bytes] = None,
        headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        raise_error: Optional[bool] = False,
        timeout: httpx._types.TimeoutTypes = 60,
    ) -> httpx.Response:
        """
        Performs POST request to server with given parameters using httpx.

        :param method: Method name.
        :type method: str
        :param json: Dictionary to send in the body of request.
        :type json: dict, optional
        :param content: Bytes with data content or dictionary with params.
        :type content: bytes or dict, optional
        :param files: Files to send in the body of request.
        :type files: dict, optional
        :param params: URL query parameters.
        :type params: str, bytes, optional
        :param headers: Custom headers to include in the request.
        :type headers: dict, optional
        :param retries: The number of attempts to connect to the server.
        :type retries: int, optional
        :param raise_error: Define, if you'd like to raise error if connection is failed.
        :type raise_error: bool, optional
        :param timeout: Overall timeout for the request.
        :type timeout: float, optional
        :return: Response object
        :rtype: :class:`httpx.Response`
        """
        self._set_client()

        if retries is None:
            retries = self.retry_count

        url = self.api_server_address + "/v3/" + method
        logger.trace(f"POST {url}")

        if headers is None:
            headers = self.headers.copy()
        else:
            headers = {**self.headers, **headers}

        for retry_idx in range(retries):
            response = None
            try:
                response = self.httpx_client.post(
                    url,
                    content=content,
                    files=files,
                    json=json,
                    params=params,
                    headers=headers,
                    timeout=timeout,
                )
                if response.status_code != httpx.codes.OK:
                    if not self._version_check_completed:
                        self._check_version()
                    Api._raise_for_status_httpx(response)
                return response
            except (httpx.RequestError, httpx.HTTPStatusError) as exc:
                if (
                    isinstance(exc, httpx.HTTPStatusError)
                    and response.status_code == 400
                    and self.token is None
                ):
                    self.logger.warning(
                        "API_TOKEN env variable is undefined. See more: https://developer.supervisely.com/getting-started/basics-of-authentication"
                    )
                if raise_error:
                    raise exc
                else:
                    process_requests_exception(
                        self.logger,
                        exc,
                        method,
                        url,
                        verbose=True,
                        swallow_exc=True,
                        sleep_sec=min(self.retry_sleep_sec * (2**retry_idx), 60),
                        response=response,
                        retry_info={"retry_idx": retry_idx + 1, "retry_limit": retries},
                    )
            except Exception as exc:
                process_unhandled_request(self.logger, exc)
        raise httpx.RequestError(
            f"Retry limit exceeded ({url})",
            request=getattr(response, "request", None),
        )

    def get_httpx(
        self,
        method: str,
        params: httpx._types.QueryParamTypes,
        retries: Optional[int] = None,
        use_public_api: Optional[bool] = True,
        timeout: httpx._types.TimeoutTypes = 60,
    ) -> httpx.Response:
        """
        Performs GET request to server with given parameters.

        :param method: Method name.
        :type method: str
        :param params: URL query parameters.
        :type params: httpx._types.QueryParamTypes
        :param retries: The number of attempts to connect to the server.
        :type retries: int, optional
        :param use_public_api: Define if public API should be used. Default is True.
        :type use_public_api: bool, optional
        :param timeout: Overall timeout for the request.
        :type timeout: float, optional
        :return: Response object
        :rtype: :class:`Response<Response>`
        """
        self._set_client()

        if retries is None:
            retries = self.retry_count

        url = self.api_server_address + "/v3/" + method
        if use_public_api is False:
            url = os.path.join(self.server_address, method)
        logger.trace(f"GET {url}")

        if isinstance(params, Dict):
            request_params = {**params, **self.additional_fields}
        else:
            request_params = params

        for retry_idx in range(retries):
            response = None
            try:
                response = self.httpx_client.get(
                    url,
                    params=request_params,
                    headers=self.headers,
                    timeout=timeout,
                )
                if response.status_code != httpx.codes.OK:
                    Api._raise_for_status_httpx(response)
                return response
            except (httpx.RequestError, httpx.HTTPStatusError) as exc:
                if (
                    isinstance(exc, httpx.HTTPStatusError)
                    and response.status_code == 400
                    and self.token is None
                ):
                    self.logger.warning(
                        "API_TOKEN env variable is undefined. See more: https://developer.supervisely.com/getting-started/basics-of-authentication"
                    )
                process_requests_exception(
                    self.logger,
                    exc,
                    method,
                    url,
                    verbose=True,
                    swallow_exc=True,
                    sleep_sec=min(self.retry_sleep_sec * (2**retry_idx), 60),
                    response=response,
                    retry_info={"retry_idx": retry_idx + 2, "retry_limit": retries},
                )
            except Exception as exc:
                process_unhandled_request(self.logger, exc)

    def stream(
        self,
        method: str,
        method_type: Literal["GET", "POST"],
        data: Union[bytes, Dict],
        headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        range_start: Optional[int] = None,
        range_end: Optional[int] = None,
        raise_error: Optional[bool] = False,
        chunk_size: int = 8192,
        use_public_api: Optional[bool] = True,
        timeout: httpx._types.TimeoutTypes = 60,
    ) -> Generator:
        """
        Performs streaming GET or POST request to server with given parameters.
        Multipart is not supported.

        :param method: Method name for the request.
        :type method: str
        :param method_type: Request type ('GET' or 'POST').
        :type method_type: str
        :param data: Bytes with data content or dictionary with params.
        :type data: bytes or dict
        :param headers: Custom headers to include in the request.
        :type headers: dict, optional
        :param retries: The number of retry attempts.
        :type retries: int, optional
        :param range_start: Start byte position for streaming.
        :type range_start: int, optional
        :param range_end: End byte position for streaming.
        :type range_end: int, optional
        :param raise_error: If True, raise raw error if the request fails.
        :type raise_error: bool, optional
        :param chunk_size: Size of the chunks to stream.
        :type chunk_size: int, optional
        :param use_public_api: Define if public API should be used.
        :type use_public_api: bool, optional
        :param timeout: Overall timeout for the request.
        :type timeout: float, optional
        :return: Generator object.
        :rtype: :class:`Generator`
        """
        self._set_client()

        if retries is None:
            retries = self.retry_count

        url = self.api_server_address + "/v3/" + method
        if not use_public_api:
            url = os.path.join(self.server_address, method)

        if headers is None:
            headers = self.headers.copy()
        else:
            headers = {**self.headers, **headers}

        logger.trace(f"{method_type} {url}")

        if isinstance(data, (bytes, Generator)):
            content = data
            json_body = None
            params = None
        elif isinstance(data, Dict):
            json_body = {**data, **self.additional_fields}
            content = None
            params = None
        else:
            params = data
            content = None
            json_body = None

        if range_start is not None or range_end is not None:
            headers["Range"] = f"bytes={range_start or ''}-{range_end or ''}"
            logger.debug(f"Setting Range header: {headers['Range']}")

        for retry_idx in range(retries):
            total_streamed = 0
            try:
                if method_type == "POST":
                    response = self.httpx_client.stream(
                        method_type,
                        url,
                        content=content,
                        json=json_body,
                        params=params,
                        headers=headers,
                        timeout=timeout,
                    )

                elif method_type == "GET":
                    response = self.httpx_client.stream(
                        method_type,
                        url,
                        params=json_body or params,
                        headers=headers,
                        timeout=timeout,
                    )
                else:
                    raise NotImplementedError(
                        f"Unsupported method type: {method_type}. Supported types: 'GET', 'POST'"
                    )

                with response as resp:
                    expected_size = int(resp.headers.get("content-length", 0))
                    if resp.status_code not in [
                        httpx.codes.OK,
                        httpx.codes.PARTIAL_CONTENT,
                    ]:
                        if not self._version_check_completed:
                            self._check_version()
                        Api._raise_for_status_httpx(resp)

                    hhash = resp.headers.get("x-content-checksum-sha256", None)
                    try:
                        for chunk in resp.iter_raw(chunk_size):
                            yield chunk, hhash
                            total_streamed += len(chunk)
                    except Exception as e:
                        raise RetryableRequestException(repr(e))

                    if expected_size != 0 and total_streamed != expected_size:
                        raise ValueError(
                            f"Streamed size does not match the expected: {total_streamed} != {expected_size}"
                        )
                    logger.trace(f"Streamed size: {total_streamed}, expected size: {expected_size}")
                    return
            except (httpx.RequestError, httpx.HTTPStatusError, RetryableRequestException) as e:
                if (
                    isinstance(e, httpx.HTTPStatusError)
                    and resp.status_code == 400
                    and self.token is None
                ):
                    self.logger.warning(
                        "API_TOKEN env variable is undefined. See more: https://developer.supervisely.com/getting-started/basics-of-authentication"
                    )
                retry_range_start = total_streamed + (range_start or 0)
                if total_streamed != 0:
                    retry_range_start += 1
                headers["Range"] = f"bytes={retry_range_start}-{range_end or ''}"
                logger.debug(f"Setting Range header {headers['Range']} for retry")
                if raise_error:
                    raise e
                else:
                    process_requests_exception(
                        self.logger,
                        e,
                        method,
                        url,
                        verbose=True,
                        swallow_exc=True,
                        sleep_sec=min(self.retry_sleep_sec * (2**retry_idx), 60),
                        response=locals().get("resp"),
                        retry_info={"retry_idx": retry_idx + 1, "retry_limit": retries},
                    )
            except Exception as e:
                process_unhandled_request(self.logger, e)
        raise httpx.RequestError(
            message=f"Retry limit exceeded ({url})",
            request=resp.request if locals().get("resp") else None,
        )

    async def post_async(
        self,
        method: str,
        json: Dict = None,
        content: Union[str, bytes, Iterable[bytes], AsyncIterable[bytes]] = None,
        files: Union[Mapping] = None,
        params: Union[str, bytes] = None,
        headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        raise_error: Optional[bool] = False,
        timeout: httpx._types.TimeoutTypes = 60,
    ) -> httpx.Response:
        """
        Performs POST request to server with given parameters using httpx.

        :param method: Method name.
        :type method: str
        :param json: Dictionary to send in the body of request.
        :type json: dict, optional
        :param content: Bytes with data content or dictionary with params.
        :type content: bytes or dict, optional
        :param files: Files to send in the body of request.
        :type files: dict, optional
        :param params: URL query parameters.
        :type params: str, bytes, optional
        :param headers: Custom headers to include in the request.
        :type headers: dict, optional
        :param retries: The number of attempts to connect to the server.
        :type retries: int, optional
        :param raise_error: Define, if you'd like to raise error if connection is failed.
        :type raise_error: bool, optional
        :param timeout: Overall timeout for the request.
        :type timeout: float, optional
        :return: Response object
        :rtype: :class:`httpx.Response`
        """
        self._set_async_client()

        if retries is None:
            retries = self.retry_count

        url = self.api_server_address + "/v3/" + method
        logger.trace(f"POST {url}")

        if headers is None:
            headers = self.headers.copy()
        else:
            headers = {**self.headers, **headers}

        for retry_idx in range(retries):
            response = None
            try:
                response = await self.async_httpx_client.post(
                    url,
                    content=content,
                    files=files,
                    json=json,
                    params=params,
                    headers=headers,
                    timeout=timeout,
                )
                if response.status_code != httpx.codes.OK:
                    if not self._version_check_completed:
                        self._check_version()
                    Api._raise_for_status_httpx(response)
                return response
            except (httpx.RequestError, httpx.HTTPStatusError) as exc:
                if (
                    isinstance(exc, httpx.HTTPStatusError)
                    and response.status_code == 400
                    and self.token is None
                ):
                    self.logger.warning(
                        "API_TOKEN env variable is undefined. See more: https://developer.supervisely.com/getting-started/basics-of-authentication"
                    )
                if raise_error:
                    raise exc
                else:
                    await process_requests_exception_async(
                        self.logger,
                        exc,
                        method,
                        url,
                        verbose=True,
                        swallow_exc=True,
                        sleep_sec=min(self.retry_sleep_sec * (2**retry_idx), 60),
                        response=response,
                        retry_info={"retry_idx": retry_idx + 1, "retry_limit": retries},
                    )
            except Exception as exc:
                process_unhandled_request(self.logger, exc)
        raise httpx.RequestError(
            f"Retry limit exceeded ({url})",
            request=getattr(response, "request", None),
        )

    async def stream_async(
        self,
        method: str,
        method_type: Literal["GET", "POST"],
        data: Union[bytes, Dict],
        headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        range_start: Optional[int] = None,
        range_end: Optional[int] = None,
        chunk_size: int = 8192,
        use_public_api: Optional[bool] = True,
        timeout: httpx._types.TimeoutTypes = 60,
        **kwargs,
    ) -> AsyncGenerator:
        """
        Performs asynchronous streaming GET or POST request to server with given parameters.
        Yield chunks of data and hash of the whole content to check integrity of the data stream.

        :param method: Method name for the request.
        :type method: str
        :param method_type: Request type ('GET' or 'POST').
        :type method_type: str
        :param data: Bytes with data content or dictionary with params.
        :type data: bytes or dict
        :param headers: Custom headers to include in the request.
        :type headers: dict, optional
        :param retries: The number of retry attempts.
        :type retries: int, optional
        :param range_start: Start byte position for streaming.
        :type range_start: int, optional
        :param range_end: End byte position for streaming.
        :type range_end: int, optional
        :param chunk_size: Size of the chunk to read from the stream. Default is 8192.
        :type chunk_size: int, optional
        :param use_public_api: Define if public API should be used.
        :type use_public_api: bool, optional
        :param timeout: Overall timeout for the request.
        :type timeout: float, optional
        :return: Async generator object.
        :rtype: :class:`AsyncGenerator`
        """
        self._set_async_client()

        if retries is None:
            retries = self.retry_count

        url = self.api_server_address + "/v3/" + method
        if not use_public_api:
            url = os.path.join(self.server_address, method)
        logger.trace(f"{method_type} {url}")

        if headers is None:
            headers = self.headers.copy()
        else:
            headers = {**self.headers, **headers}

        params = kwargs.get("params", None)
        if "content" in kwargs or "json_body" in kwargs:
            content = kwargs.get("content", None)
            json_body = kwargs.get("json_body", None)
        else:
            if isinstance(data, (bytes, Generator)):
                content = data
                json_body = None
            elif isinstance(data, Dict):
                json_body = {**data, **self.additional_fields}
                content = None
            else:
                raise ValueError("Data should be either bytes or dict")

        if range_start is not None or range_end is not None:
            headers["Range"] = f"bytes={range_start or ''}-{range_end or ''}"
            logger.debug(f"Setting Range header: {headers['Range']}")

        for retry_idx in range(retries):
            total_streamed = 0
            try:
                if method_type == "POST":
                    response = self.async_httpx_client.stream(
                        method_type,
                        url,
                        content=content,
                        json=json_body,
                        headers=headers,
                        timeout=timeout,
                        params=params,
                    )
                elif method_type == "GET":
                    response = self.async_httpx_client.stream(
                        method_type,
                        url,
                        content=content,
                        json=json_body,
                        headers=headers,
                        timeout=timeout,
                        params=params,
                    )
                else:
                    raise NotImplementedError(
                        f"Unsupported method type: {method_type}. Supported types: 'GET', 'POST'"
                    )

                async with response as resp:
                    expected_size = int(resp.headers.get("content-length", 0))
                    if resp.status_code not in [
                        httpx.codes.OK,
                        httpx.codes.PARTIAL_CONTENT,
                    ]:
                        if not self._version_check_completed:
                            self._check_version()
                        Api._raise_for_status_httpx(resp)

                    # received hash of the content to check integrity of the data stream
                    hhash = resp.headers.get("x-content-checksum-sha256", None)
                    try:
                        async for chunk in resp.aiter_raw(chunk_size):
                            yield chunk, hhash
                            total_streamed += len(chunk)
                    except Exception as e:
                        raise RetryableRequestException(repr(e))

                    if expected_size != 0 and total_streamed != expected_size:
                        raise ValueError(
                            f"Streamed size does not match the expected: {total_streamed} != {expected_size}"
                        )
                    logger.trace(f"Streamed size: {total_streamed}, expected size: {expected_size}")
                    return
            except (httpx.RequestError, httpx.HTTPStatusError, RetryableRequestException) as e:
                if (
                    isinstance(e, httpx.HTTPStatusError)
                    and resp.status_code == 400
                    and self.token is None
                ):
                    self.logger.warning(
                        "API_TOKEN env variable is undefined. See more: https://developer.supervisely.com/getting-started/basics-of-authentication"
                    )
                retry_range_start = total_streamed + (range_start or 0)
                if total_streamed != 0:
                    retry_range_start += 1
                headers["Range"] = f"bytes={retry_range_start}-{range_end or ''}"
                logger.debug(f"Setting Range header {headers['Range']} for retry")
                await process_requests_exception_async(
                    self.logger,
                    e,
                    method,
                    url,
                    verbose=True,
                    swallow_exc=True,
                    sleep_sec=min(self.retry_sleep_sec * (2**retry_idx), 60),
                    response=locals().get("resp"),
                    retry_info={"retry_idx": retry_idx + 1, "retry_limit": retries},
                )
            except Exception as e:
                process_unhandled_request(self.logger, e)
        raise httpx.RequestError(
            message=f"Retry limit exceeded ({url})",
            request=resp.request if locals().get("resp") else None,
        )

    def _set_async_client(self):
        """
        Set async httpx client with HTTP/2 if it is not set yet.
        """
        if self.async_httpx_client is None:
            self.async_httpx_client = httpx.AsyncClient(http2=True)

    def _set_client(self):
        """
        Set sync httpx client with HTTP/2 if it is not set yet.
        """
        if self.httpx_client is None:
            self.httpx_client = httpx.Client(http2=True)

    def get_default_semaphore(self) -> asyncio.Semaphore:
        """
        Get default global API semaphore for async requests.
        If the semaphore is not set, it will be initialized.

        During initialization, the semaphore size will be set from the environment variable SUPERVISELY_ASYNC_SEMAPHORE.
        If the environment variable is not set, the default value will be set based on the server address.
        Depending on the server address, the semaphore size will be set to 10 for HTTPS and 5 for HTTP.

        :return: Semaphore object.
        :rtype: :class:`asyncio.Semaphore`
        """
        if self._semaphore is None:
            self._initialize_semaphore()
        return self._semaphore

    def _initialize_semaphore(self):
        """
        Initialize the semaphore for async requests.

        If the environment variable SUPERVISELY_ASYNC_SEMAPHORE is set, create a semaphore with the given value.

        Otherwise, create a semaphore with a default value:

            - If server supports HTTPS, create a semaphore with value 10.
            - If server supports HTTP, create a semaphore with value 5.
        """
        semaphore_size = sly_env.semaphore_size()
        if semaphore_size is not None:
            self._semaphore = asyncio.Semaphore(semaphore_size)
            logger.debug(
                f"Setting global API semaphore size to {semaphore_size} from environment variable"
            )
        else:
            if not self._skip_https_redirect_check:
                self._check_https_redirect()
            if self.server_address.startswith("https://"):
                size = 10
                if "app.supervise" in self.server_address:
                    size = 7
                logger.debug(f"Setting global API semaphore size to {size} for HTTPS")
            else:
                size = 5
                logger.debug(f"Setting global API semaphore size to {size} for HTTP")
            self._semaphore = asyncio.Semaphore(size)

    def set_semaphore_size(self, size: int = None):
        """
        Set the global API semaphore with the given size. Will replace the existing semaphore.
        If the size is not set, will set from the environment variable SUPERVISELY_ASYNC_SEMAPHORE.
        If the environment variable is not set, will set the default value.

        :param size: Size of the semaphore.
        :type size: int, optional
        """
        if size is not None:
            self._semaphore = asyncio.Semaphore(size)
        else:
            self._initialize_semaphore()

    @property
    def semaphore(self) -> asyncio.Semaphore:
        """
        Get the global API semaphore for async requests.

        :return: Semaphore object.
        :rtype: :class:`asyncio.Semaphore`
        """
        return self._semaphore



from __future__ import annotations

import traceback
from functools import wraps
from typing import Any, List, Literal, NamedTuple, Tuple, Union

from supervisely import logger
from supervisely.app.content import DataJson, StateJson
from supervisely.app.fastapi.utils import run_sync
from supervisely.app.fastapi.websocket import WebsocketManager
from supervisely.app.widgets import Widget
from supervisely.imaging.color import hex2rgb, rgb2hex

"""
size1 = 10
x1 = list(range(size1))
y1 = np.random.randint(low=10, high=148, size=size1).tolist()
s1 = [{"x": x, "y": y} for x, y in zip(x1, y1)]

size2 = 30
x2 = list(range(size2))
y2 = np.random.randint(low=0, high=300, size=size2).tolist()
s2 = [{"x": x, "y": y} for x, y in zip(x2, y2)]

chart = sly.app.widgets.Apexchart(
    series=[{"name": "Max", "data": s1}, {"name": "Denis", "data": s2}],
    options={
        "chart": {"type": "line", "zoom": {"enabled": False}},
        "dataLabels": {"enabled": False},
        # "stroke": {"curve": "straight"},
        "stroke": {"curve": "smooth", "width": 2},
        "title": {"text": "Product Trends by Month", "align": "left"},
        "grid": {"row": {"colors": ["#f3f3f3", "transparent"], "opacity": 0.5}},
        "xaxis": {"type": "category"},
    },
    type="line",
)
"""


class Apexchart(Widget):
    class Routes:
        CLICK = "chart_clicked_cb"

    class ClickedDataPoint(NamedTuple):
        series_index: int
        series_name: str
        data_index: int
        data: dict
        x: Any
        y: Any

    def __init__(
        self,
        series: List[dict],
        options: dict,
        type: str,
        height: Union[int, str] = "300",
        sly_options: dict = {},
    ):
        self._series = series
        self._options = options
        self._type = type
        self._height = height
        self._click_handled = False
        self._sly_options = sly_options
        self._data_type: Literal["dict", "tuple"] = self._options.get("data_type", "dict")
        super().__init__(file_path=__file__)

    def get_json_data(self):
        return {
            "series": self._series,
            "options": self._options,
            "type": self._type,
            "height": self._height,
            "sly_options": self._sly_options,
        }

    def get_json_state(self):
        return {"clicked_value": None}

    def get_clicked_value(self):
        return StateJson()[self.widget_id]["clicked_value"]

    def get_clicked_datapoint(self) -> Apexchart.ClickedDataPoint:
        value = self.get_clicked_value()
        series_index = value["seriesIndex"]
        data_index = value["dataPointIndex"]
        if series_index == -1 and data_index != -1:
            # zero point (0,0) click
            series_index = 0
        if series_index == -1 or data_index == -1:
            return
        series_name = self._series[series_index]["name"]
        data = self._series[series_index]["data"][data_index]
        res = Apexchart.ClickedDataPoint(
            series_index, series_name, data_index, data, data["x"], data["y"]
        )
        return res

    def click(self, func):
        route_path = self.get_route_path(Apexchart.Routes.CLICK)
        server = self._sly_app.get_server()

        self._click_handled = True

        @server.post(route_path)
        def _click():
            res = self.get_clicked_datapoint()
            if res is not None:
                try:
                    return func(res)
                except Exception as e:
                    logger.error(traceback.format_exc(), exc_info=True, extra={"exc_str": str(e)})
                    raise e

        return _click

    def _wrap_xy(self, px, py) -> Union[dict, tuple]:
        return {"x": px, "y": py} if self._data_type == "dict" else (px, py)

    def add_series(self, name: str, x: list, y: list, send_changes=True):
        if len(x) != len(y):
            raise ValueError(f"Lists x and y have different lenght, {len(x)} != {len(y)}")

        data = [self._wrap_xy(px, py) for px, py in zip(x, y)]
        series = {"name": name, "data": data}
        self._series.append(series)
        self.update_data()
        if send_changes:
            DataJson().send_changes()

    def set_title(self, title: str, send_changes=True):
        if self._options.get("title") is None:
            self._options["title"] = {}
        self._options["title"]["text"] = title
        self.update_data()
        if send_changes:
            DataJson().send_changes()

    def set_series(self, series: list, send_changes=True):
        self._series = series
        self.update_data()
        if send_changes:
            DataJson().send_changes()

    def set_colors(self, colors: List[str or List[int]], send_changes: bool = True) -> None:
        """
        Set colors for every series in the chart.

        :param colors: Set new colors for every series as a color name, rgb, or HEX. Example: `{'title' : ['red', 'rgb(0,255,0)', '#0000FF']}` or `{'title' : [[255,0,0], [0,255,0], [0,0,255]]}`.
        :type colors: List[str | List[int]]
        :param send_changes: Send changes to the chart. Defaults to True.
        :type send_changes: bool
        """
        clrs = colors
        if isinstance(colors[0], list):  # f.e. [255, 0, 0]
            clrs = [rgb2hex(c) for c in colors]

        self._options["colors"] = clrs
        self.update_data()
        if send_changes:
            DataJson().send_changes()

    def _wrap_to_series_data(
        self, data: Union[List[tuple], List[dict], tuple, dict]
    ) -> List[Union[dict, tuple]]:
        if isinstance(data, list):
            if len(data) == 0:
                return []
            if isinstance(data[0], dict):
                return [self._wrap_xy(dct["x"], dct["y"]) for dct in data]
            return [self._wrap_xy(*tpl) for tpl in data]

        else:  # if single datapoint
            if isinstance(data, dict):
                return [self._wrap_xy(data["x"], data["y"])]
            return [self._wrap_xy(*data)]

    def add_to_series(
        self,
        name_or_id: Union[str, int],
        data: Union[List[tuple], List[dict], tuple, dict],
        send_changes=True,
    ) -> None:
        if isinstance(name_or_id, int):
            series_id = name_or_id
        else:
            series_id, _ = self.get_series_by_name(name_or_id)

        data_list = self._wrap_to_series_data(data)
        self._series[series_id]["data"].extend(data_list)
        DataJson()[self.widget_id]["series"] = self._series

        if send_changes:
            DataJson().send_changes()
            self._force_series_update()

    def get_series_by_name(self, name: str) -> Tuple[int, dict]:
        series_list = DataJson()[self.widget_id]["series"]
        series_id, series_data = next(
            ((i, series) for i, series in enumerate(series_list) if series["name"] == name),
            (None, None),
        )
        return series_id, series_data

    def _force_series_update(self) -> None:
        # necessary for the force rendering of data in apexchart
        run_sync(
            WebsocketManager().broadcast(
                {
                    "runAction": {
                        "action": f"sly-apexchart-{self.widget_id}.update-series",
                        "payload": {},
                    }
                }
            )
        )



from typing import Any, Callable, Dict, List, Optional

import supervisely as sly
from supervisely.app import DataJson
from supervisely.app.content import StateJson
from supervisely.app.widgets import Widget
from supervisely.geometry.cuboid_3d import Cuboid3d
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.point_3d import Point3d
from supervisely.geometry.pointcloud import Pointcloud
from supervisely.sly_logger import logger

type_to_zmdi_icon = {
    sly.AnyGeometry: "zmdi zmdi-shape",
    sly.Rectangle: "zmdi zmdi-crop-din",  # "zmdi zmdi-square-o"
    # sly.Polygon: "icons8-polygon",  # "zmdi zmdi-edit"
    sly.Polygon: "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAABmJLR0QA/wD/AP+gvaeTAAAB6klEQVRYhe2Wuy8EURTGf+u5VESNXq2yhYZCoeBv8RcI1i6NVUpsoVCKkHjUGlFTiYb1mFmh2MiKjVXMudmb3cPOzB0VXzKZm5k53/nmvO6Ff4RHD5AD7gFP1l3Kd11AHvCBEpAVW2esAvWmK6t8l1O+W0lCQEnIJoAZxUnzNQNkZF36jrQjgoA+uaciCgc9VaExBOyh/6WWAi1VhbjOJ4FbIXkBtgkK0BNHnYqNKUIPeBPbKyDdzpld5T6wD9SE4AwYjfEDaXFeFzE/doUWuhqwiFsOCwqv2hV2lU/L+sHBscGTxdvSFVoXpAjCZdauMHVic6ndl6U1VBsJCFhTeNUU9IiIEo3qvQYGHAV0AyfC5wNLhKipXuBCjA5wT8WxcM1FMRoBymK44CjAE57hqIazwCfwQdARcXa3UXHuRXVucIjb7jYvNkdxBZg0TBFid7PQTRAtX2xOiXkuMAMqYwkIE848rZFbjyNAmw9bIeweaZ2A5TgC7PnwKkTPtN+cTOrsyN3FEWAjRTAX6sA5ek77gSL6+WHZVQDAIHAjhJtN78aAS3lXAXYIivBOnCdyOAUYB6o0xqsvziry7FLE/Cp20cNcJEjDr8MUmVOVRzkVN+Nd7vZGVXXgiwxtPiRS5WFhz4fEq/zv4AvToMn7vCn3eAAAAABJRU5ErkJggg==",
    sly.Bitmap: "zmdi zmdi-brush",
    sly.Polyline: "zmdi zmdi-gesture",
    sly.Point: "zmdi zmdi-dot-circle-alt",
    sly.Cuboid: "zmdi zmdi-ungroup",  #
    sly.GraphNodes: "zmdi zmdi-grain",
    Cuboid3d: "zmdi zmdi-codepen",
    Pointcloud: "zmdi zmdi-cloud-outline",  # "zmdi zmdi-border-clear"
    sly.MultichannelBitmap: "zmdi zmdi-layers",  # "zmdi zmdi-collection-item"
    Point3d: "zmdi zmdi-filter-center-focus",  # "zmdi zmdi-select-all"
}


class ClassesTable(Widget):
    """ClassesTable widget in Supervisely allows users to display all classes from given project in a table format.

    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/tables/classestable>`_
        (including screenshots and examples).

    :param project_meta: Project meta object from which classes will be taken.
    :type project_meta: sly.ProjectMeta
    :param project_id: Project id from which classes will be taken.
    :type project_id: int
    :param project_fs: Project object from which classes will be taken.
    :type project_fs: sly.Project
    :param allowed_types: List of allowed geometry types to be displayed in table.
    :type allowed_types: List[Geometry]
    :param selectable: If True, user can select classes from table.
    :type selectable: bool
    :param disabled: If True, the elements in the table will be disabled.
    :type disabled: bool
    :param widget_id: Unique widget identifier.
    :type widget_id: str
    :raises ValueError: If both project_id and project_fs parameters are provided.

    :Usage example:
    .. code-block:: python

        from supervisely.app.widgets import ClassesTable

        classes_table = ClassesTable(project_id=123, selectable=True)
    """

    class Routes:
        CLASS_SELECTED = "class_selected_cb"

    def __init__(
        self,
        project_meta: Optional[sly.ProjectMeta] = None,
        project_id: Optional[int] = None,
        project_fs: Optional[sly.Project] = None,
        allowed_types: Optional[List[Geometry]] = None,
        selectable: Optional[bool] = True,
        disabled: Optional[bool] = False,
        widget_id: Optional[str] = None,
        dataset_ids: Optional[List[int]] = None,
    ):
        if project_id is not None and project_fs is not None:
            raise ValueError(
                "You can not provide both project_id and project_fs parameters to Classes Table widget."
            )
        self._table_data = []
        self._columns = []
        self._changes_handled = False
        self._global_checkbox = False
        self._checkboxes = []
        self._selectable = selectable
        self._selection_disabled = disabled
        self._loading = False
        self._allowed_types = allowed_types if allowed_types is not None else []
        if project_id is not None:
            self._api = sly.Api()
        else:
            self._api = None
        self._project_id = project_id
        if project_id is not None:
            if project_meta is not None:
                logger.warn(
                    "Both parameters project_id and project_meta were provided to ClassesTable widget. Project meta classes taken from remote project and project_meta parameter is ignored."
                )
            project_meta = sly.ProjectMeta.from_json(self._api.project.get_meta(project_id))
        self._project_fs = project_fs
        if project_fs is not None:
            if project_meta is not None:
                logger.warn(
                    "Both parameters project_fs and project_meta were provided to ClassesTable widget. Project meta classes taken from project_fs.meta and project_meta parameter is ignored."
                )
            project_meta = project_fs.meta

        self._project_meta = project_meta
        self._dataset_ids = dataset_ids
        if project_meta is not None:
            self._update_meta(project_meta=project_meta)
        super().__init__(widget_id=widget_id, file_path=__file__)

    def value_changed(self, func: Callable[[List[str]], Any]) -> Callable[[], None]:
        """Decorator for the function to be called when the value of the widget changes.

        :param func: Function to be called when the value of the widget changes.
        :type func: Callable[[List[str]], Any]
        :return: Decorated function.
        :rtype: Callable[[], None]
        """
        route_path = self.get_route_path(ClassesTable.Routes.CLASS_SELECTED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _value_changed():
            res = self.get_selected_classes()
            func(res)

        return _value_changed

    def _update_meta(self, project_meta: sly.ProjectMeta) -> None:
        columns = ["class", "shape"]
        stats = None
        data_to_show = []
        for obj_class in project_meta.obj_classes:
            if len(self._allowed_types) == 0 or obj_class.geometry_type in self._allowed_types:
                data_to_show.append(obj_class.to_json())

        if self._project_id is not None:
            stats = self._api.project.get_stats(self._project_id)
            project_info = self._api.project.get_info_by_id(self._project_id)
            if project_info.type == str(sly.ProjectType.IMAGES):
                columns.extend(["images count", "labels count"])
            elif project_info.type == str(sly.ProjectType.VIDEOS):
                columns.extend(["videos count", "figures count"])
            elif project_info.type in [
                str(sly.ProjectType.POINT_CLOUDS),
                str(sly.ProjectType.POINT_CLOUD_EPISODES),
            ]:
                columns.extend(["pointclouds count", "figures count"])
            elif project_info.type == str(sly.ProjectType.VOLUMES):
                columns.extend(["volumes count", "figures count"])

            class_items = {}
            for item in stats["images"]["objectClasses"]:
                if self._dataset_ids is not None:
                    items_count = 0
                    for cls_dataset_stats in item["datasets"]:
                        if cls_dataset_stats["id"] in self._dataset_ids:
                            items_count += cls_dataset_stats["count"]
                    class_items[item["objectClass"]["name"]] = items_count
                else:
                    class_items[item["objectClass"]["name"]] = item["total"]

            class_objects = {}
            for item in stats["objects"]["items"]:
                if self._dataset_ids is not None:
                    items_count = 0
                    for cls_dataset_stats in item["datasets"]:
                        if cls_dataset_stats["id"] in self._dataset_ids:
                            items_count += cls_dataset_stats["count"]
                    class_objects[item["objectClass"]["name"]] = items_count
                else:
                    class_objects[item["objectClass"]["name"]] = item["total"]
            for obj_class in data_to_show:
                obj_class["itemsCount"] = class_items[obj_class["title"]]
                obj_class["objectsCount"] = class_objects[obj_class["title"]]

        elif self._project_fs is not None:
            project_stats = self._project_fs.get_classes_stats()

            if type(self._project_fs) == sly.Project:
                columns.extend(["images count", "labels count"])

            elif type(self._project_fs) == sly.VideoProject:
                columns.extend(["videos count", "objects count", "figures count"])

            elif type(self._project_fs) in [sly.PointcloudProject, sly.PointcloudEpisodeProject]:
                columns.extend(["pointclouds count", "objects count", "figures count"])

            elif type(self._project_fs) == sly.VolumeProject:
                columns.extend(["volumes count", "objects count", "figures count"])

            for obj_class in data_to_show:
                obj_class["itemsCount"] = project_stats["items_count"][obj_class["title"]]
                obj_class["objectsCount"] = project_stats["objects_count"][obj_class["title"]]
                if type(self._project_fs) != sly.Project:
                    obj_class["figuresCount"] = project_stats["figures_count"][obj_class["title"]]

        columns = [col.upper() for col in columns]
        if data_to_show:
            table_data = []
            if self._project_id is not None or self._project_fs is not None:
                data_to_show = sorted(
                    data_to_show, key=lambda line: line["objectsCount"], reverse=True
                )
            for line in data_to_show:
                table_line = []
                icon = type_to_zmdi_icon[sly.AnyGeometry]
                for geo_type, icon_text in type_to_zmdi_icon.items():
                    geo_type: Geometry
                    if geo_type.geometry_name() == line["shape"]:
                        icon = icon_text
                        break
                if line["shape"] == "graph":
                    line["shape"] = "graph (keypoints)"
                table_line.extend(
                    [
                        {
                            "name": "CLASS",
                            "data": line["title"],
                            "color": line["color"],
                        },
                        {"name": "SHAPE", "data": line["shape"], "icon": icon},
                    ]
                )
                if "itemsCount" in line.keys():
                    table_line.append({"name": "ITEMS COUNT", "data": line["itemsCount"]})
                if "objectsCount" in line.keys():
                    table_line.append({"name": "OBJECTS COUNT", "data": line["objectsCount"]})
                if "figuresCount" in line.keys():
                    table_line.append({"name": "FIGURES COUNT", "data": line["figuresCount"]})
                table_data.append(table_line)
            self._table_data = table_data
            self._columns = columns
            self._checkboxes = [False] * len(table_data)
            self._global_checkbox = False
        else:
            self._table_data = []
            self._columns = []
            self._checkboxes = []
            self._global_checkbox = False

    def read_meta(self, project_meta: sly.ProjectMeta) -> None:
        """Read project meta and update table data.

        :param project_meta: Project meta object from which classes will be taken.
        :type project_meta: sly.ProjectMeta
        """
        self.loading = True
        self._project_fs = None
        self._project_id = None
        self._project_meta = project_meta
        self.clear_selection()
        self._update_meta(project_meta=project_meta)

        DataJson()[self.widget_id]["table_data"] = self._table_data
        DataJson()[self.widget_id]["columns"] = self._columns
        DataJson().send_changes()
        StateJson()[self.widget_id]["checkboxes"] = self._checkboxes
        StateJson()[self.widget_id]["global_checkbox"] = self._global_checkbox
        StateJson().send_changes()
        self.loading = False

    def read_project(self, project_fs: sly.Project) -> None:
        """Read local project and update table data.

        :param project_fs: Project object from which classes will be taken.
        :type project_fs: sly.Project
        """
        self.loading = True
        self._project_fs = project_fs
        self._project_id = None
        self._project_meta = project_fs.meta
        self.clear_selection()
        self._update_meta(project_meta=project_fs.meta)

        DataJson()[self.widget_id]["table_data"] = self._table_data
        DataJson()[self.widget_id]["columns"] = self._columns
        DataJson().send_changes()
        StateJson()[self.widget_id]["checkboxes"] = self._checkboxes
        StateJson()[self.widget_id]["global_checkbox"] = self._global_checkbox
        StateJson().send_changes()
        self.loading = False

    def read_project_from_id(
        self, project_id: int, dataset_ids: Optional[List[int]] = None
    ) -> None:
        """Read remote project by id and update table data.

        :param project_id: Project id from which classes will be taken.
        :type project_id: int
        :param dataset_ids: List of dataset ids to filter classes.
        :type dataset_ids: Optional[List[int]]
        """
        self.loading = True
        self._project_fs = None
        self._project_id = project_id
        if dataset_ids is not None:
            self._dataset_ids = dataset_ids
        if self._api is None:
            self._api = sly.Api()
        project_meta = sly.ProjectMeta.from_json(self._api.project.get_meta(project_id))
        self._project_meta = project_meta
        self.clear_selection()
        self._update_meta(project_meta=project_meta)

        DataJson()[self.widget_id]["table_data"] = self._table_data
        DataJson()[self.widget_id]["columns"] = self._columns
        DataJson().send_changes()
        StateJson()[self.widget_id]["checkboxes"] = self._checkboxes
        StateJson()[self.widget_id]["global_checkbox"] = self._global_checkbox
        StateJson().send_changes()
        self.loading = False

    def get_json_data(self) -> Dict[str, Any]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - table_data: List of dictionaries with table data.
            - columns: List of column names.
            - loading: If True, the widget is in loading state.
            - disabled: If True, the elements in the table will be disabled.
            - selectable: If True, user can select classes from table.

        :return: Dictionary with widget data.
        :rtype: Dict[str, Any]
        """
        return {
            "table_data": self._table_data,
            "columns": self._columns,
            "loading": self._loading,
            "disabled": self._selection_disabled,
            "selectable": self._selectable,
        }

    @property
    def allowed_types(self) -> List[Geometry]:
        """Returns list of allowed geometry types to be displayed in table.

        :return: List of allowed geometry types to be displayed in table.
        :rtype: List[Geometry]
        """
        return self._allowed_types

    @property
    def project_id(self) -> int:
        """Returns project id from which classes was taken.

        :return: Project id from which classes was taken.
        :rtype: int
        """
        return self._project_id

    @property
    def project_fs(self) -> int:
        """Returns project object from which classes was taken.

        :return: Project object from which classes was taken.
        :rtype: sly.Project
        """
        return self._project_fs

    @property
    def loading(self) -> bool:
        """Returns True if the widget is in loading state.

        :return: True if the widget is in loading state.
        :rtype: bool
        """
        return self._loading

    @property
    def project_meta(self) -> bool:
        """Returns project meta object from which classes was taken.

        :return: Project meta object from which classes was taken.
        :rtype: sly.ProjectMeta
        """
        return self._project_meta

    @loading.setter
    def loading(self, value: bool) -> None:
        """Sets loading state of the widget.

        :param value: Loading state of the widget.
        :type value: bool
        """
        self._loading = value
        DataJson()[self.widget_id]["loading"] = self._loading
        DataJson().send_changes()

    def get_json_state(self) -> Dict[str, Any]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - global_checkbox: State of global checkbox.
            - checkboxes: List of checkboxes states.

        :return: Dictionary with widget state.
        :rtype: Dict[str, Any]
        """
        return {
            "global_checkbox": self._global_checkbox,
            "checkboxes": self._checkboxes,
        }

    def get_selected_classes(self) -> List[str]:
        """Returns list of selected classes.

        :return: List of selected classes.
        :rtype: List[str]
        """
        classes = []
        checkboxes = StateJson()[self.widget_id]["checkboxes"]
        for i, line in enumerate(self._table_data):
            if len(checkboxes) == 0:
                checkboxes = [False] * len(self._table_data)
            if i >= len(checkboxes):
                continue
            if checkboxes[i]:
                for col in line:
                    if col["name"] == "CLASS":
                        classes.append(col["data"])
        return classes

    def clear_selection(self) -> None:
        """Clears selection of classes."""
        self._global_checkbox = False
        self._checkboxes = [False] * len(self._table_data)
        StateJson()[self.widget_id]["global_checkbox"] = self._global_checkbox
        StateJson()[self.widget_id]["checkboxes"] = self._checkboxes
        StateJson().send_changes()

    def set_project_meta(self, project_meta: sly.ProjectMeta) -> None:
        """Sets project meta object from which classes will be taken.

        :param project_meta: Project meta object from which classes will be taken.
        :type project_meta: sly.ProjectMeta
        """
        self._update_meta(project_meta)
        self._project_meta = project_meta
        self.update_data()
        DataJson().send_changes()

    def select_classes(self, classes: List[str]) -> None:
        """Selects classes in the table from given list.

        :param classes: List of classes to be selected.
        :type classes: List[str]
        """
        self._global_checkbox = False
        self._checkboxes = [False] * len(self._table_data)

        project_classes = []
        for i, line in enumerate(self._table_data):
            for col in line:
                if col["name"] == "CLASS":
                    project_classes.append(col["data"])

        for i, cls_name in enumerate(project_classes):
            if cls_name in classes:
                self._checkboxes[i] = True
        self._global_checkbox = all(self._checkboxes)
        StateJson()[self.widget_id]["global_checkbox"] = self._global_checkbox
        StateJson()[self.widget_id]["checkboxes"] = self._checkboxes
        StateJson().send_changes()

    def select_all(self) -> None:
        """Selects all classes in the table."""
        self._global_checkbox = True
        self._checkboxes = [True] * len(self._table_data)
        StateJson()[self.widget_id]["global_checkbox"] = self._global_checkbox
        StateJson()[self.widget_id]["checkboxes"] = self._checkboxes
        StateJson().send_changes()

    def set_dataset_ids(self, dataset_ids: List[int]) -> None:
        """Sets dataset ids to filter classes.

        :param dataset_ids: List of dataset ids to filter classes.
        :type dataset_ids: List[int]
        """
        selected_classes = self.get_selected_classes()
        self._dataset_ids = dataset_ids
        self._update_meta(self._project_meta)
        self.update_data()
        self.select_classes(selected_classes)
